# cython: language_level=3
from cython cimport view

from pyglui.ui import get_opensans_font_path, get_roboto_font_path

from pyglui.cygl cimport glew as gl
from pyglui.cygl.utils cimport RGBA
# pyfontstash needs to be import AFTER pyglui.cygl.glew
from pyglui.pyfontstash cimport fontstash as fs

from os import path

__version__ = "1.31.0"


# legacy push and pop

def push_view(int w, int h):
    pass

def pop_view():
    pass


cdef class Bar_Graph:
    cdef fs.Context glfont
    cdef double[::1] data
    cdef public float avg, bar_width, min_val, max_val
    cdef float scale
    cdef int idx,d_len
    cdef int x, y, win_height, win_width
    cdef basestring label
    cdef int s_idx,s_size
    cdef object data_source
    cdef public RGBA color

    def __cinit__(self,int data_points = 25,float min_val = 0, float max_val = 100):
        self.data = view.array(shape=(data_points,), itemsize=sizeof(double), format="d")
        self.d_len = data_points
        self.label = 'Title %0.2f units'
        self.bar_width = 4
        self.min_val = min_val
        self.max_val = max_val
        self.scale = 1

        self.glfont = fs.Context()
        self.glfont.add_font('opensans', get_opensans_font_path())
        self.glfont.set_size(int(18*self.scale))
        self.color = RGBA(0.5, 0.8, 0.75,.5)

    def __init__(self,int data_points = 25,float min_val = 0, float max_val = 100):
        cdef int x
        for x in range(data_points):
            self.data[x] = 0
        self.avg = 0
        self.data_source = lambda: 0

    def adjust_window_size(self, w, h):
        self.win_width = w
        self.win_height = h

    property label:
        def __get__(self):
            return self.label
        def __set__(self,new_label):
            self.label = new_label

    property scale:
        def __get__(self):
            return self.scale

        def __set__(self,val):
            self.scale = val
            self.glfont.set_size(int(18*self.scale))

    property pos:
        def __get__(self):
            return self.x,self.y

        def __set__(self,val):
            self.x,self.y = val

    property update_rate:
        def __set__(self,r):
            self.s_size = r
            self.s_idx = 0

        def __get__(self):
            return self.s_size

    property update_fn:
        def __set__(self,fn):
            self.data_source = fn

    def add(self,double val):
        self.s_idx = (self.s_idx +1) %self.s_size
        if self.s_idx == 0:
            self.data[self.idx] = val
            self.idx = (self.idx +1) %self.d_len
            #very rough avg estimation
            self.avg += (val-self.avg)*(5./self.d_len)

    def update(self):
        cdef float val
        self.s_idx = (self.s_idx +1) %self.s_size
        if self.s_idx == 0:
            val = self.data_source()
            self.data[self.idx] = val
            self.idx = (self.idx +1) %self.d_len
            #very rough avg estimation
            self.avg += (val-self.avg)*(2./self.d_len)


    def draw(self):
        cdef int i
        cdef float x=0
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        gl.glOrtho(0, self.win_width, self.win_height, 0, -1, 1)

        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        gl.glTranslatef(self.x*self.scale,self.y*self.scale,0)
        #scale such that a bar at max val is 100px high
        gl.glPushMatrix()
        gl.glScalef(1,-30.*self.scale/self.max_val,1)
        gl.glTranslatef(0,self.min_val*self.scale,0)

        ##draw background
        #gl.glColor4f(.0,.0,.0,.2)
        #gl.glBegin(gl.GL_POLYGON)
        #gl.glVertex3f(0,0,0.0)
        #gl.glVertex3f(100,0,0.0)
        #gl.glVertex3f(100,100,0.0)
        #gl.glVertex3f(0,100,0.0)
        #gl.glEnd()
        #draw bars
        gl.glLineWidth(self.bar_width*self.scale)
        gl.glColor4f(self.color.r,self.color.g,self.color.b,self.color.a)
        gl.glBegin(gl.GL_LINES)

        #if self.s_size:
        #    x -= self.s_idx/float(self.s_size) * 4

        for i in range(self.idx,self.d_len):
            gl.glVertex3f(x*self.scale,0,0)
            gl.glVertex3f(x*self.scale,self.data[i],0)
            x += self.bar_width

        for i in range(self.idx):
            gl.glVertex3f(x*self.scale,0,0)
            gl.glVertex3f(x*self.scale,self.data[i],0)
            x += self.bar_width

        gl.glEnd()
        gl.glPopMatrix()


        self.glfont.draw_text(0,-3,self.label%self.avg)
        gl.glPopMatrix()

        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPopMatrix()
        gl.glMatrixMode(gl.GL_MODELVIEW)



cdef class Line_Graph:
    cdef fs.Context glfont
    cdef double[::1] data
    cdef public float avg,bar_width,min_val, max_val
    cdef float scale
    cdef int idx,d_len
    cdef int x, y, win_height, win_width
    cdef basestring label
    cdef int s_idx,s_size
    cdef object data_source
    cdef public RGBA color


    def __cinit__(self,int data_points = 50,float min_val = 0, float max_val = 100):
        self.data = view.array(shape=(data_points,), itemsize=sizeof(double), format="d")
        self.d_len = data_points
        self.label = 'Title %0.2f units'
        self.bar_width = 2
        self.min_val = min_val
        self.max_val = max_val
        self.scale = 1.

        self.glfont = fs.Context()
        self.glfont.add_font('opensans',get_opensans_font_path())
        self.glfont.set_size(int(18*self.scale))
        self.glfont.set_align(fs.FONS_ALIGN_LEFT | fs.FONS_ALIGN_MIDDLE)
        self.color = RGBA(0.5, 0.8, 0.75,.5)

    def __init__(self,int data_points = 25,float min_val = 0, float max_val = 100):
        cdef int x
        for x in range(data_points):
            self.data[x] = 0
        self.avg = 0
        self.data_source = lambda: 0

    def adjust_window_size(self, w, h):
        self.win_width = w
        self.win_height = h

    property scale:
        def __get__(self):
            return self.scale

        def __set__(self,val):
            self.scale = val
            self.glfont.set_size(int(18*self.scale))

    property label:
        def __get__(self):
            return self.label
        def __set__(self,new_label):
            self.label = new_label

    property pos:
        def __get__(self):
            return self.x,self.y

        def __set__(self,val):
            self.x,self.y = val

    property update_rate:
        def __set__(self,r):
            self.s_size = r
            self.s_idx = 0

        def __get__(self):
            return self.s_size

    property update_fn:
        def __set__(self,fn):
            self.data_source = fn


    def add(self,double val):
        self.s_idx = (self.s_idx +1) %self.s_size
        if self.s_idx == 0:
            self.data[self.idx] = val
            self.idx = (self.idx +1) %self.d_len
            #very rough avg estimation
            self.avg += (val-self.avg)*(5./self.d_len)

    def update(self):
        cdef float val
        self.s_idx = (self.s_idx +1) %self.s_size
        if self.s_idx == 0:
            val = self.data_source()
            self.data[self.idx] = val
            self.idx = (self.idx +1) %self.d_len
            #very rough avg estimation
            self.avg += (val-self.avg)*(2./self.d_len)


    def draw(self):
        cdef int i
        cdef float x=0

        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        gl.glOrtho(0, self.win_width, self.win_height, 0, -1, 1)


        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        gl.glTranslatef(self.x*self.scale,self.y*self.scale,0)
        #scale such that a bar at max val is 100px high
        gl.glPushMatrix()
        gl.glScalef(1,-100.*self.scale/self.max_val,1)
        gl.glTranslatef(0,self.min_val*self.scale,0)


        gl.glLineWidth(self.bar_width*self.scale)
        gl.glColor4f(self.color.r,self.color.g,self.color.b,self.color.a)
        gl.glBegin(gl.GL_LINE_STRIP)


        for i in range(self.idx,self.d_len):
            gl.glVertex3f(x*self.scale,self.data[i],0)
            x += self.bar_width

        for i in range(self.idx):
            gl.glVertex3f(x*self.scale,self.data[i],0)
            x += self.bar_width

        gl.glEnd()
        gl.glPopMatrix()
        self.glfont.set_color_float(self.color[:])
        self.glfont.draw_text(x +10 ,-self.data[i],self.label%self.avg)
        gl.glPopMatrix()

        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPopMatrix()
        gl.glMatrixMode(gl.GL_MODELVIEW)




cdef class Averaged_Value:
    cdef fs.Context glfont
    cdef double[::1] data
    cdef basestring label
    cdef int idx,d_len
    cdef public float avg
    cdef int s_idx,s_size
    cdef int x, y, font_size
    cdef float scale
    cdef object data_source
    cdef public RGBA color

    def __cinit__(self, int data_points=25, int font_size=18):
        self.data = view.array(shape=(data_points,), itemsize=sizeof(double), format="d")
        self.scale = 1.
        self.d_len = data_points
        self.label = 'Title %0.2f units'
        self.font_size = font_size
        self.glfont = fs.Context()
        self.glfont.add_font('opensans',get_opensans_font_path())
        self.glfont.set_size(int(self.font_size*self.scale))
        self.glfont.set_align(fs.FONS_ALIGN_LEFT | fs.FONS_ALIGN_MIDDLE)
        self.color = RGBA(1.,1.,1.,1.)

    def __init__(self, int data_points=25,int font_size=18):
        cdef int x
        for x in range(data_points):
            self.data[x] = 0
        self.avg = 0
        self.data_source = lambda: 0

    property label:
        def __get__(self):
            return self.label
        def __set__(self,new_label):
            self.label = new_label

    property pos:
        def __get__(self):
            return self.x,self.y
        def __set__(self,val):
            self.x,self.y = val

    property update_rate:
        def __set__(self,r):
            self.s_size = r
            self.s_idx = 0
        def __get__(self):
            return self.s_size

    property scale:
        def __get__(self):
            return self.scale

        def __set__(self,val):
            self.scale = val
            self.glfont.set_size(int(self.font_size*self.scale))

    def set_text_align(self,v_align='left',h_align='top'):
        v_align = {'left':fs.FONS_ALIGN_LEFT,'center':fs.FONS_ALIGN_CENTER,'right':fs.FONS_ALIGN_RIGHT}[v_align]
        h_align = {'top':fs.FONS_ALIGN_TOP,'middle':fs.FONS_ALIGN_MIDDLE,'bottom':fs.FONS_ALIGN_BOTTOM}[h_align]
        self.glfont.set_align(v_align | h_align)


    def add(self,double val):
        self.s_idx = (self.s_idx +1) %self.s_size
        if self.s_idx == 0:
            self.data[self.idx] = val
            self.idx = (self.idx +1) %self.d_len
            #very rough avg estimation
            self.avg += (val-self.avg)*(5./self.d_len)

    def update(self):
        cdef float val
        self.s_idx = (self.s_idx +1) %self.s_size
        if self.s_idx == 0:
            val = self.data_source()
            self.data[self.idx] = val
            self.idx = (self.idx +1) %self.d_len
            #very rough avg estimation
            self.avg += (val-self.avg)*(2./self.d_len)

    def draw(self):
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPushMatrix()
        self.glfont.set_color_float(self.color[:])
        self.glfont.draw_text(self.x*self.scale,self.y*self.scale,self.label%self.avg)
        gl.glPopMatrix()



########## UI_element ##########
# Base class
# see 'design_params.pxi' for design parameters
cdef class UI_element:
    '''
    The base class for all UI elements.
    '''
    cdef float _order
    cdef readonly object uid
    cdef public FitBox outline
    cdef basestring _label
    cdef bint _read_only


    cpdef sync(self):
        global should_redraw
        pass

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        pass

    cpdef draw_overlay(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        pass

    cpdef handle_input(self,Input new_input,bint visible, bint parent_read_only = False):
        if not (self._read_only or parent_read_only):
            global should_redraw
            pass

    cpdef pre_handle_input(self,Input new_input):
        # if your element needs to catch input event in front of everybody else:
        # add yourself to new_input.active_ui_elements during handle_input
        # pre_handle input will be called in the next frame with new input.
        # add yourself again in handle input if you need to stay in active_ui_elements
        pass

    cpdef precompute(self,FitBox parent):
        self.outline.compute(parent)

    property width:
        def __get__(self):
            return self.outline.size.x

    property height:
        def __get__(self):
            return self.outline.size.y

    property read_only:
        def __get__(self):
            return self._read_only
        def __set__(self,bint val):
            if self._read_only != val:
                self._read_only = val
                global should_redraw
                should_redraw = True

    property label:
        def __get__(self):
            return self._label
        def __set__(self,basestring val):
            if self._label != val:
                self._label = val
                global should_redraw
                should_redraw = True

    property order:
        def __get__(self):
            return self._order
        def __set__(self,float order):
            if self._order != order:
                self._order = order
                global should_redraw
                should_redraw = True


########## Slider ##########
#    +--------------------------------+
#    | Label  Input Text              |
#    | ------------------O----------- |
#    +--------------------------------+

cdef class Slider(UI_element):
    cdef public float minimum,maximum,step
    cdef public FitBox field
    cdef bint selected
    cdef Vec2 slider_pos
    cdef Synced_Value sync_val
    cdef int steps
    cdef RGBA line_default_color, line_highlight_color, text_color, button_color, button_selected_color, button_shadow_color,step_color
    cdef Slider_Text_Input label_field

    def __cinit__(self,str attribute_name, object attribute_context = None, label = None, min = 0, max = 100, step = 0,setter= None,getter= None):

        self.label_field = Slider_Text_Input(attribute_name, attribute_context,
                                             label=label or attribute_name,
                                             setter=setter, getter=getter)
        self.label_field.validator = self.validate
        self.uid = id(self)
        self.sync_val = self.label_field.sync_val
        self.step = abs(step)
        self.minimum = min
        if self.step:
            self.maximum = (((max-min+.0000001)//self.step))*self.step+min # + 0.0000001 becasue of floating point rounding issues.
        else:
            self.maximum = max
        self.outline = FitBox(Vec2(0,0),Vec2(0,slider_outline_size_y)) # we only fix the height
        self.field = FitBox(Vec2(outline_padding,outline_padding),Vec2(-outline_padding,-outline_padding))
        self.slider_pos = Vec2(0,slider_label_org_y)
        self.selected = False
        self._read_only = False
        if self.step:
            self.steps = int((self.maximum-self.minimum)/float(step))
        else:
            self.steps = 0
        self.line_default_color = RGBA(*slider_line_color_default)
        self.line_highlight_color = RGBA(*slider_line_color_highlight)
        self.text_color = RGBA(*color_text_default)
        self.button_color = RGBA(*color_default)
        self.button_selected_color = RGBA(*color_selected)
        self.button_shadow_color = RGBA(*color_shadow)
        self.step_color = RGBA(*slider_color_step)
        self.display_format = '%0.2f'

    def __init__(self,str attribute_name, object attribute_context = None,label = None, min = 0, max = 100, step = 1,setter= None,getter= None):
        self.sync()
        if not isinstance(self.sync_val.value, (float,int) ):
            raise Exception('Slider values should be float or int type. "%s" is of type %s'%(self.sync_val.value,type(self.sync_val.value)))


    cpdef sync(self):
        self.sync_val.sync()

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        #update appearance:
        self.outline.compute(parent)
        self.field.compute(self.outline)

        # map slider value
        self.slider_pos.x = clampmap(self.sync_val.value,self.minimum,self.maximum,0,self.field.size.x)
        self.slider_pos.y = slider_handle_org_y * ui_scale
        #self.outline.sketch()
        #self.field.sketch()

        # read only rendering rules
        # cdef tuple line_default_color, line_highlight_color, text_color, button_color, button_shadow_color
        if self._read_only or parent_read_only:
            self.line_default_color = RGBA(*slider_line_color_default_read_only)
            self.line_highlight_color = RGBA(*slider_line_color_highlight_read_only)
            self.text_color = RGBA(*color_text_read_only)
            self.button_color = RGBA(*color_default_read_only)
            self.button_shadow_color = RGBA(*color_shadow_read_only)
        else:
            self.line_default_color = RGBA(*slider_line_color_default)
            self.line_highlight_color = RGBA(*slider_line_color_highlight)
            self.text_color = RGBA(*color_text_default)
            self.button_color = RGBA(*color_default)
            self.button_shadow_color = RGBA(*color_shadow)

        gl.glPushMatrix()
        gl.glTranslatef(self.field.org.x,self.field.org.y,0)


        rect_corners(Vec2(self.slider_pos.x, self.slider_pos.y - 1. * ui_scale),
                     Vec2(self.field.size.x, self.slider_pos.y +  1. * ui_scale),
                     self.line_default_color)
        rect_corners(Vec2(0., self.slider_pos.y - 1. * ui_scale),
                     Vec2(self.slider_pos.x, self.slider_pos.y +  1. * ui_scale),
                     self.line_highlight_color)

        cdef float step_pixel_size,x
        if self.steps>1:
            step_pixel_size = lmap(self.minimum+self.step,self.minimum,self.maximum,0,self.field.size.x)
            if step_pixel_size >= slider_button_size*ui_scale:
                step_marks = [(x*step_pixel_size,self.slider_pos.y) for x in range(self.steps+1)]
                utils.draw_points(step_marks,size=slider_step_mark_size*ui_scale, color=self.step_color)

        if self.selected:
            utils.draw_points((self.slider_pos,),size=(slider_button_size_selected+slider_button_shadow)*ui_scale, color=self.button_shadow_color,sharpness=shadow_sharpness)
            utils.draw_points((self.slider_pos,),size=slider_button_size_selected*ui_scale, color=self.button_selected_color)
        else:
            utils.draw_points((self.slider_pos,),size=(slider_button_size+slider_button_shadow)+ui_scale, color=self.button_shadow_color,sharpness=shadow_sharpness)
            utils.draw_points((self.slider_pos,),size=slider_button_size*ui_scale, color=self.button_color)

        gl.glPopMatrix()
        self.label_field.draw(self.outline, nested=True, parent_read_only=(self._read_only or parent_read_only))



    cpdef handle_input(self,Input new_input,bint visible,bint parent_read_only = False):
        if not (self._read_only or parent_read_only):
            global should_redraw

            if self.selected and new_input.dm:
                val = clampmap(new_input.m.x-self.field.org.x,0,self.field.size.x,self.minimum,self.maximum)
                #conserve some spcial types.
                if isinstance(self.sync_val.value,int):
                    self.sync_val.value = int(step(val,self.minimum,self.maximum,self.step))
                else:
                    self.sync_val.value = step(val,self.minimum,self.maximum,self.step)

                should_redraw = True

            for b in new_input.buttons[:]:#list copy for remove to work
                if b[1] == 1 and visible:
                    if mouse_over_center(Vec2(self.field.org.x + self.field.size.x / 2.,
                                              self.slider_pos.y + self.field.org.y),
                                         self.field.size.x,self.field.size.y/2.,new_input.m):
                        new_input.buttons.remove(b) # the slider should catch the event (unlike other elements)
                        self.selected = True
                        should_redraw = True
                if self.selected and b[1] == 0:
                    self.selected = False
                    should_redraw = True

            self.label_field.handle_input(new_input, visible, False)

    cpdef validate(self, val):
        return step(clamp(val, self.minimum, self.maximum), self.minimum, self.maximum, self.step)

    @property
    def display_format(self):
        return self.label_field.display_format

    @display_format.setter
    def display_format(self, val):
        self.label_field.display_format = val


########## Switch ##########
#
#   +--------------------------------+
#   | Label                        O |
#   +--------------------------------+


cdef class Switch(UI_element):
    cdef public FitBox field,button
    cdef bint selected
    cdef int on_val,off_val
    cdef Synced_Value sync_val
    cdef RGBA text_color, button_color_on, button_color_off, button_shadow_color, button_selected_color

    def __cinit__(self,str attribute_name, object attribute_context = None, on_val=True, off_val=False, label=None, setter=None, getter=None):
        self.uid = id(self)
        self._label = label or attribute_name
        self.sync_val = Synced_Value(attribute_name,attribute_context,getter,setter)
        self.on_val = on_val
        self.off_val = off_val
        self.outline = FitBox(Vec2(0,0),Vec2(0,switch_outline_size_y)) # we only fix the height
        self.field = FitBox(Vec2(outline_padding,outline_padding),Vec2(-outline_padding,-outline_padding))
        self.button = FitBox(Vec2(-switch_button_size-x_spacer,0),Vec2(switch_button_size,switch_button_size))
        self.selected = False
        # rendering variables
        self.text_color = RGBA(*color_text_default)
        self.button_color_on = RGBA(*color_on)
        self.button_color_off = RGBA(*color_default)
        self.button_shadow_color = RGBA(*color_shadow)
        self.button_selected_color = RGBA(*color_selected)


    def __init__(self,str attribute_name, object attribute_context = None,label = None, on_val = True, off_val = False ,setter= None,getter= None):
        pass


    cpdef sync(self):
        self.sync_val.sync()

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        #update appearance
        self.outline.compute(parent)
        self.field.compute(self.outline)
        self.button.compute(self.field)

        # read only rendering rules
        if self._read_only or parent_read_only:
            self.text_color = RGBA(*color_text_read_only)
            self.button_color_on = RGBA(*color_on_read_only)
            self.button_color_off = RGBA(*color_default_read_only)
            self.button_shadow_color = RGBA(*color_shadow_read_only)
        else:
            self.text_color = RGBA(*color_text_default)
            self.button_color_on = RGBA(*color_on)
            self.button_color_off = RGBA(*color_default)
            self.button_shadow_color = RGBA(*color_shadow)


        if self.sync_val.value == self.on_val:
            # on state
            utils.draw_points(((self.button.center),),size=(switch_button_size_on+switch_button_shadow)*ui_scale, color=self.button_shadow_color,sharpness=shadow_sharpness)
            utils.draw_points(((self.button.center),),size=switch_button_size_on*ui_scale, color=self.button_color_on)
        elif self.selected:
            utils.draw_points(((self.button.center),),size=(switch_button_size_selected+switch_button_shadow)*ui_scale, color=self.button_shadow_color,sharpness=shadow_sharpness)
            utils.draw_points(((self.button.center),),size=switch_button_size_selected*ui_scale, color=self.button_selected_color)
        else:
            # off state
            utils.draw_points(((self.button.center),),size=(switch_button_size+switch_button_shadow)*ui_scale, color=self.button_shadow_color,sharpness=shadow_sharpness)
            utils.draw_points(((self.button.center),),size=switch_button_size*ui_scale, color=self.button_color_off)


        gl.glPushMatrix()
        gl.glTranslatef(self.field.org.x,self.field.org.y,0)

        glfont.push_state()
        glfont.set_color_float(self.text_color[:])

        glfont.draw_limited_text(x_spacer,0,self._label,self.field.size.x-(switch_button_size_on+switch_button_shadow))

        glfont.pop_state()
        gl.glPopMatrix()

    cpdef handle_input(self,Input new_input,bint visible,bint parent_read_only = False):
        if not (self._read_only or parent_read_only):
            global should_redraw

            for b in new_input.buttons:
                if visible and self.button.mouse_over(new_input.m):
                    if b[1] == 1:
                        self.selected = True
                        should_redraw = True
                if self.selected and b[1] == 0 and (self.sync_val.value == self.on_val):
                    self.sync_val.value = self.off_val
                    self.selected = False
                    should_redraw = True
                if self.selected and b[1] == 0 and (self.sync_val.value == self.off_val):
                    self.sync_val.value = self.on_val
                    self.selected = False
                    should_redraw = True




########## Selector ##########
#
#   +--------------------------------+
#   |       +----------------------+ |
#   | Label | Selection          v | |
#   |       +----------------------+ |
#   +--------------------------------+


cdef class Selector(UI_element):
    cdef public FitBox field
    cdef FitBox select_field
    cdef object selection, selection_labels
    cdef Synced_Value sync_val
    cdef int selection_idx
    cdef bint selected
    cdef RGBA text_color, triangle_color
    cdef object selection_getter

    def __cinit__(self,str attribute_name, object attribute_context = None, selection = [], labels=None, label=None, setter=None, getter=None, selection_getter = None):
        # NOTE: implementing a custom selection_getter can lead to race-conditions. Make
        # sure that the current set value is always contained in the returned selections
        # from selection_getter.

        self.uid = id(self)
        self._label = label or attribute_name

        self.text_color = RGBA(*color_text_default)
        self.triangle_color = RGBA(*selector_triangle_color_default)

        def default_selection_getter():
            def to_str(obj):
                if isinstance(obj,basestring):
                    return obj
                else:
                    return str(obj)

            return selection,(labels or [to_str(s) for s in selection])

        self.selection_getter = selection_getter or default_selection_getter

        self.selection,self.selection_labels = self.selection_getter()

        self.sync_val = Synced_Value(attribute_name,attribute_context,getter,setter,self._on_change)

        self.outline = FitBox(Vec2(0,0),Vec2(0,selector_outline_size_y)) # we only fix the height
        self.field = FitBox(Vec2(outline_padding,outline_padding),Vec2(-outline_padding,-outline_padding))
        self.select_field = FitBox(Vec2(x_spacer,0),Vec2(0,0))

    def __init__(self,*args,**kwargs):
        pass

    cpdef sync(self):
        self.sync_val.sync()

    def _on_change(self,new_value):
        self.selection,self.selection_labels = self.selection_getter()
        try:
            self.selection_idx = list(self.selection).index(new_value)
        except ValueError:
            ##we could throw and error here or ignore
            ##but for now we just add the new object to the selection
            #self.selection.append(new_value)
            #self.selection_labels.append(str(new_value))
            raise ValueError("Synced value '%s' is not part of selection."%str(new_value))


    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only = False):

        #update appearance
        self.outline.compute(parent)
        self.field.compute(self.outline)
        self.select_field.compute(self.field)

        # read only rendering rules
        if self._read_only or parent_read_only:
            self.text_color = RGBA(*color_text_read_only)
            self.triangle_color = RGBA(*selector_triangle_color_read_only)
        else:
            self.text_color = RGBA(*color_text_default)
            if self.selected:
                self.triangle_color = RGBA(*color_on)
            else:
                self.triangle_color = RGBA(*selector_triangle_color_default)

        gl.glPushMatrix()
        gl.glTranslatef(self.field.org.x,self.field.org.y,0)
        glfont.push_state()
        glfont.set_color_float(self.text_color[:])

        cdef float label_text_space = glfont.draw_text(x_spacer,0,self._label)
        glfont.pop_state()
        gl.glPopMatrix()

        self.select_field.org.x += label_text_space + x_spacer * ui_scale
        self.select_field.size.x  = max(0.0,self.select_field.size.x-label_text_space - x_spacer * ui_scale)
        rect_outline(self.select_field.org, self.select_field.size, 2.*ui_scale, self.triangle_color)
        #self.select_field.sketch()
        # line(self.select_field.org+Vec2(0,self.select_field.size.y),self.select_field.org+self.select_field.size,color=self.triangle_color)


        gl.glPushMatrix()
        gl.glTranslatef(self.select_field.org.x,self.select_field.org.y,0)

        glfont.push_state()
        glfont.set_color_float(self.text_color[:])

        if self.selected:
            for y in range(len(self.selection)):
                glfont.draw_limited_text(x_spacer,y*line_height*ui_scale,self.selection_labels[y],self.select_field.size.x-x_spacer)
        else:
            glfont.draw_limited_text(x_spacer,0,self.selection_labels[self.selection_idx],self.select_field.size.x-x_spacer-self.select_field.size.y)
            triangle_h(self.select_field.size-Vec2(self.select_field.size.y,self.select_field.size.y),
                    Vec2(self.select_field.size.y,self.select_field.size.y),
                    self.triangle_color)
        glfont.pop_state()
        gl.glPopMatrix()

    cpdef handle_input(self,Input new_input,bint visible,bint parent_read_only = False):
        if not (self._read_only or parent_read_only):
            global should_redraw

            for b in new_input.buttons:
                if visible and self.select_field.mouse_over(new_input.m):
                    if b[1] == 0:
                        should_redraw = True
                        if not self.selected:
                            self.init_selection()
                        else:
                            self.finish_selection(mouse_y = new_input.m.y-self.select_field.org.y)

    cdef init_selection(self):
        self.selected = True
        self.selection,self.selection_labels = self.selection_getter()
        #blow up the menu to fit the open selector field
        cdef h = line_height * len(self.selection_labels)
        h+= self.field.design_org.y - self.field.design_size.y #double neg
        h+= self.select_field.design_org.y - self.select_field.design_size.y
         #double neg
        self.outline.design_size.y = h


    cdef finish_selection(self,float mouse_y):
        clicked_idx = int(mouse_y/(self.select_field.size.y/float(len(self.selection))) )
        #just for sanity
        clicked_idx = int(clamp(clicked_idx,0,len(self.selection)-1))

        self.sync_val.value = self.selection[clicked_idx]

        # Sync value. Will update self.selection_idx if the value actually changed.
        self.sync_val.sync()

        #make the outline small again
        cdef h = line_height
        h+= self.field.design_org.y - self.field.design_size.y #double neg
        h+= self.select_field.design_org.y - self.select_field.design_size.y #double neg
        self.outline.design_size.y = h

        self.selected = False

########## TextInput ##########
#
#   +--------------------------------+
#   | Label  Input text              |
#   +--------------------------------+

cdef class Text_Input(UI_element):
    '''
    Text input field.
    '''
    cdef FitBox field, textfield
    cdef bint selected,highlight, catch_input
    cdef readonly Synced_Value sync_val
    cdef unicode preview
    cdef int caret,start_char_idx,end_char_idx,start_highlight_idx
    cdef RGBA text_color, text_input_highlight_color, text_input_line_highlight_color
    cdef object data_type
    cdef double t_long_press, t_double_click

    def __cinit__(self,str attribute_name, object attribute_context = None,label = None,setter= None,getter= None):
        self.uid = id(self)
        self._label = label or attribute_name
        self.sync_val = Synced_Value(attribute_name,attribute_context,getter,setter)
        self.outline = FitBox(Vec2(0,0),Vec2(0,text_input_outline_size_y)) # we only fix the height
        self.field = FitBox(Vec2(outline_padding,outline_padding),Vec2(-outline_padding,-outline_padding))
        self.textfield = FitBox(Vec2(x_spacer,0),Vec2(0,0))
        self.selected = False
        self.highlight = False
        self.preview = u'' #only used when self.selected==True
        self.caret = 0 #only used when self.selected==True
        self.start_char_idx = 0
        self.end_char_idx = self.caret
        self.start_highlight_idx = 0
        self.text_color = RGBA(*color_text_default)
        self.text_input_highlight_color = RGBA(*text_input_highlight_color)
        self.text_input_line_highlight_color = RGBA(*text_input_line_highlight_color)
        self.data_type = type(self.sync_val.value)
        self.catch_input = True
        self.t_long_press = 0.0 #timer used for long press
        self.t_double_click = 0.0 #timer used for t_double_click

    def __init__(self,str attribute_name, object attribute_context = None,label = None,setter= None,getter= None):
        pass


    cpdef sync(self):
        self.sync_val.sync()

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only = False):

        # read only rendering rules
        if self._read_only or parent_read_only:
            self.text_color = RGBA(*color_text_read_only)
        else:
            self.text_color = RGBA(*color_text_default)

        #update appearance:
        self.outline.compute(parent)
        self.field.compute(self.outline)
        self.textfield.compute(self.field)

        glfont.push_state()
        glfont.set_color_float(self.text_color[:])

        gl.glPushMatrix()
        gl.glTranslatef(self.field.org.x,self.field.org.y,0)
        dx = glfont.draw_text(x_spacer,0,self._label)
        gl.glPopMatrix()

        self.textfield.org.x += dx + x_spacer * ui_scale
        self.textfield.size.x -= dx + x_spacer * ui_scale

        self.draw_text_field()
        glfont.pop_state()
        # self.draw_text_selection()


    cpdef pre_handle_input(self,Input new_input):
        global should_redraw
        while new_input.keys:
            key,scancode,action,mods = new_input.keys.pop(0)
            # See file glfw.py
            # for key,scancode,action,mods definitions
            if key == 257 and action == 0: #Enter and key press
                self.finish_input()
                return
            if  key == 256 and action == 0: #ESC and key press
                self.abort_input()
                return

            # cut: mod+x (88), copy: mod+c (67)
            elif key in (67, 88) and mods & UI_MOD_KEY and action != 0:
                if self.highlight:
                    min_idx = min(self.start_highlight_idx, self.caret)
                    max_idx = max(self.start_highlight_idx, self.caret)
                    new_input.cb = self.preview[min_idx:max_idx]
                    if key == 88:
                        self.preview = self.preview[:min_idx] + self.preview[max_idx:]
                        self.caret = min_idx
                        self.highlight = False
                        should_redraw = True
            # paste: mod+v
            elif key == 86 and mods & UI_MOD_KEY and action != 0 and new_input.cb:
                if self.highlight:
                    min_idx = min(self.start_highlight_idx, self.caret)
                    max_idx = max(self.start_highlight_idx, self.caret)
                else:
                    min_idx = max_idx = self.caret

                self.preview = self.preview[:min_idx] + new_input.cb + self.preview[max_idx:]
                self.caret = min_idx + len(new_input.cb)
                self.highlight = False
                should_redraw = True

            elif key == 259 and action != 1: # Backspace and key not released (key repeat)
                if self.caret > 0 and self.highlight is False:
                    self.preview = self.preview[:self.caret-1] + self.preview[self.caret:]
                    self.caret -=1
                if self.highlight:
                    min_idx = min(self.start_highlight_idx, self.caret)
                    max_idx = max(self.start_highlight_idx, self.caret)
                    self.preview = self.preview[:min_idx] + self.preview[max_idx:]
                    self.caret = min_idx
                    self.highlight = False

                self.caret = max(0, self.caret)
                should_redraw = True

            elif key == 261 and action != 1: # Delete and key not released (key repeat)
                if self.caret < len(self.preview) and self.highlight is False:
                    self.preview = self.preview[:self.caret] + self.preview[self.caret+1:]
                if self.highlight:
                    self.preview = self.preview[:min(self.start_highlight_idx,self.caret)] + self.preview[max(self.start_highlight_idx,self.caret):]
                    self.caret = min(self.start_highlight_idx,self.caret)
                    self.highlight = False

                self.caret = min(len(self.preview), self.caret)
                should_redraw = True

            elif key == 263 and action != 1 and mods == 0: #key left and key not released without mod keys
                self.caret -=1
                self.caret = max(0,self.caret)
                self.highlight=False
                should_redraw = True

            elif key == 262 and action != 1 and mods == 0: #key right and key not released without mod keys
                self.caret +=1
                self.caret = min(len(self.preview),self.caret)
                self.highlight=False
                should_redraw = True

            elif  key == 263 and action != 1 and mods == 1: #key left and key not released with shift mod
                if self.highlight is False:
                    self.start_highlight_idx = max(0,self.caret)
                self.caret -=1
                self.caret = max(0,self.caret)
                self.highlight = True
                should_redraw = True

            elif key == 262 and action != 1 and mods == 1: #key right and key not released with shift mod
                if self.highlight is False:
                    self.start_highlight_idx = min(len(self.preview),self.caret)
                self.caret +=1
                self.caret = min(len(self.preview),self.caret)
                self.highlight = True
                should_redraw = True

            elif key == 65 and action == 0 and mods == UI_MOD_KEY: # select all
                # key a and action key press and mods are either command/super for MacOS or control for Windows
                if len(self.preview) > 0 and self.highlight is False:
                    self.start_highlight_idx = 0
                    self.caret = len(self.preview)
                    self.highlight = True
                should_redraw = True
            elif key in (268, 269) and action == 1:  # home on release
                if mods != 1:
                    self.highlight = False
                elif self.highlight is False:
                    self.start_highlight_idx = min(len(self.preview),self.caret)
                    self.highlight = True

                if key == 269:
                    self.caret = len(self.preview)
                else:
                    self.caret = 0
                should_redraw = True

        while new_input.chars:
            c = new_input.chars.pop(0)
            if self.highlight:
                # new char overwrites all highlighted chars
                self.preview = self.preview[:min(self.start_highlight_idx,self.caret)] + c + self.preview[max(self.start_highlight_idx,self.caret):]
                self.caret = min(self.start_highlight_idx+1,self.caret+1)
                self.caret = max(1,self.caret)
                self.highlight = False
            else:
                self.preview = self.preview[:self.caret] + c + self.preview[self.caret:]
                self.caret +=1

            should_redraw = True

        for b in new_input.buttons:
            if b[1] == 0 and not self.textfield.mouse_over(new_input.m):
                self.finish_input()

    cdef to_unicode(self,obj):
        if type(obj) is unicode:
            return obj
        elif type(obj) is bytes:
            return obj.decode('utf-8')
        else:
            return unicode(obj)

    cpdef handle_input(self,Input new_input,bint visible,bint parent_read_only = False):
        global should_redraw
        cdef double d_double_click = 0.5  # delay used for double click
        cdef double d_long_press = 1.0  # delay used for long press
        cdef double now = time()

        if not (self._read_only or parent_read_only):
            for b in new_input.buttons[:]:
                if self.textfield.mouse_over(new_input.m) and visible:
                    if b[1] == 1:
                        self.selected = True
                        self.highlight = False
                        self.preview = self.to_unicode(self.sync_val.value)
                        self.caret = len(self.preview)

                        mouse_x_in_text_box = new_input.m.x-x_spacer-self.textfield.org.x

                        # get caret position closest to current mouse.x
                        self.calculate_start_idx() # caret position needs to be updated relative to text offset in the case of long string of text
                        caret_positions = glfont.char_cumulative_width(x_spacer,0,self.preview[self.start_char_idx:self.caret])
                        mouse_to_caret = [abs(i-mouse_x_in_text_box) for i in caret_positions]
                        min_distance = min(mouse_to_caret)
                        self.caret = mouse_to_caret.index(min_distance)

                        should_redraw = True
                        self.t_long_press = now

                        if now - self.t_double_click > d_double_click:
                            self.highlight = False
                            self.t_double_click = now
                        elif len(self.preview) > 0:
                            self.highlight = True
                            self.start_highlight_idx = 0
                            self.caret = len(self.preview)
                            self.t_double_click = 0.

                    elif b[1] == 0:  # button release
                        if now - self.t_long_press > d_long_press:
                            # long press highlights all text
                            # similar to the behavior on a mobile device
                            self.highlight = True
                            self.start_highlight_idx = 0
                            self.caret = len(self.preview)

                        should_redraw = True

                    if self.catch_input:
                        # this is required so that we can catch mouse up behavior
                        new_input.buttons.remove(b)

        if self.selected:
            new_input.active_ui_elements.append(self)


    cdef finish_input(self):
        global should_redraw
        should_redraw = True
        self.selected = False
        self.update_input_val()

    cdef update_input_val(self):
        # turn string back into the data_type of the value in case of str always use unicode
        if isinstance(self.sync_val.value, basestring):
            typed_val = self.preview
        else:
            try:
                typed_val = self.data_type(eval(self.preview))
            except:
                #failed to convert. Ignore user input.
                return
        self.sync_val.value = typed_val


    cdef abort_input(self):
        global should_redraw
        should_redraw = True
        self.selected = False

    cdef calculate_start_idx(self):
        # clip the preview text appropriately so that it always fits within the textfield
        # make sure there is always one char before or after caret
        cdef float width
        cdef int start_char_idx, caret_x
        width = self.textfield.size.x - 2*x_spacer

        # sanity check - if the preview is shorter than the width, then start_char_idx should be 0
        if glfont.text_bounds(x_spacer,0,self.preview[self.start_char_idx:]) < width:
            self.start_char_idx = 0

        # get the position of the caret
        caret_x = glfont.text_bounds(x_spacer,0,self.preview[self.start_char_idx:self.caret])

        # scroll left:
        # if the caret is == start_char_idx subtract one from start_char_idx until idx = 0
        if self.caret == self.start_char_idx:
            # start_char_idx = len(self.preview)-glfont.get_clip_position(self.preview[::-1],width)
            self.start_char_idx = max(0,self.start_char_idx-1)

        # scroll right:
        # if the caret reaches the right bound
        if caret_x >= width:
            # find the starting idx of the left most (starting) char if >= width of textfield+padding
            # at the current caret position - overflow on the right is handled by draw_limited_text
            start_char_idx = glfont.get_first_char_idx(self.preview[:self.caret],width)
            # adding some preview after the caret by modifying the start_char
            # this may result in more than one extra char after, unless monospaced fonts are used
            self.start_char_idx = min(len(self.preview),start_char_idx)


    cdef draw_text_field(self):
        cdef float x

        if self.selected:
            self.calculate_start_idx()
            highlight_text = self.preview[self.start_char_idx:self.start_highlight_idx]

            gl.glPushMatrix()
            #then transform locally and render the UI element
            gl.glTranslatef(self.textfield.org.x,self.textfield.org.y,0)
            rect_outline(Vec2(0, 0), self.textfield.size, 2.*ui_scale, self.text_input_line_highlight_color)

            x = glfont.text_bounds(0,0,self.preview[self.start_char_idx:self.caret])+x_spacer
            # draw highlighted text if any
            if self.highlight:
               rect_corners(Vec2(x,0),Vec2(min(self.textfield.size.x-x_spacer,
                    glfont.text_bounds(0,0,highlight_text)+x_spacer),self.textfield.size.y),
                    self.text_input_highlight_color)

            glfont.draw_limited_text(x_spacer,0,self.preview[self.start_char_idx:],self.textfield.size.x-x_spacer)

            # draw the caret
            gl.glColor4f(1,1,1,.5)
            gl.glLineWidth(1)
            gl.glBegin(gl.GL_LINES)
            gl.glVertex3f(x,0,0)
            gl.glVertex3f(x,self.textfield.size.y,0)
            gl.glEnd()
            gl.glPopMatrix()

        else:
            gl.glPushMatrix()
            #then transform locally and render the UI element
            #self.textfield.sketch()
            gl.glTranslatef(self.textfield.org.x,self.textfield.org.y,0)
            rect_outline(Vec2(0, 0), self.textfield.size, 2.*ui_scale, RGBA(*color_line_default))
            glfont.draw_limited_text(x_spacer,0,self.to_unicode(self.sync_val.value),self.textfield.size.x-x_spacer)
            gl.glPopMatrix()


cdef class Slider_Text_Input(Text_Input):

    cdef public object validator
    cdef public basestring display_format

    def __cinit__(self, *args, **kwargs):
        self.validator = lambda x: x
        self.display_format = '%0.2f'

    cdef update_input_val(self):
        # turn string back into the data_type of the value in case of str always use unicode
        if isinstance(self.sync_val.value, basestring):
            typed_val = self.preview
        else:
            try:
                typed_val = self.data_type(self.validator(eval(self.preview)))
            except:
                #failed to convert. Ignore user input.
                return
        self.sync_val.value = typed_val

    cdef to_unicode(self,obj):
        if type(obj) is unicode:
            return obj
        elif type(obj) is bytes:
            return obj.decode('utf-8')
        elif isinstance(obj, float):
            return self.display_format % obj
        else:
            return unicode(obj)

########## Button ##########
#
#   +--------------------------------+
#   |                      +-------+ |
#   | Outer_Label          | Label | |
#   |                      +-------+ |
#   +--------------------------------+



cdef class Button(UI_element):
    cdef FitBox field, button
    cdef bint selected
    cdef public object function
    cdef RGBA text_color
    cdef basestring _outer_label

    def __cinit__(self,label, function, outer_label=''):
        self.uid = id(self)
        self._label = label
        self._outer_label = outer_label
        self.outline = FitBox(Vec2(0,0),Vec2(0,button_outline_size_y)) # we only fix the height
        self.field = FitBox(Vec2(0,0),Vec2(0,0))  # depends on string length
        self.button = FitBox(Vec2(0,0),Vec2(0,0))  # will be computed on demand
        self.selected = False
        self.function = function
        self.text_color = RGBA(*color_text_default)

    def __init__(self, *args, **kwargs):
        pass

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        cdef tuple text_color
        cdef RGBA bg_color
        # read only rendering rules
        if self._read_only or parent_read_only:
            self.text_color = RGBA(*button_read_only_text_color)
            bg_color = RGBA(*button_read_only_color)
        elif self.selected:
            self.text_color = RGBA(*button_active_text_color)
            bg_color = RGBA(*button_active_color)
        else:
            self.text_color = RGBA(*button_default_text_color)
            bg_color = RGBA(*button_default_color)

        #update appearance:
        self.outline.compute(parent)

        cdef float label_width = 0.
        if self._outer_label:
            label_width = glfont.text_bounds(0., 0., self._label) / ui_scale
            self.field = FitBox(Vec2(outline_padding, outline_padding),
                                Vec2(-label_width-outline_padding-3.*x_spacer, -outline_padding))
            self.button = FitBox(Vec2(-label_width-button_text_padding-button_outline_padding-2.*x_spacer, button_outline_padding),
                                 Vec2(label_width+2.*x_spacer+2*button_text_padding, -button_outline_padding))
            self.field.compute(self.outline)
        else:
            self.button = FitBox(Vec2(button_outline_padding, button_outline_padding),
                                 Vec2(-button_outline_padding, -button_outline_padding))
        self.button.compute(self.outline)

        cdef FitBox shadow = self.button.computed_copy()
        if not self.selected:
            shadow.org -= Vec2(2. * ui_scale, 2. * ui_scale)
            shadow.size += Vec2(4. * ui_scale, 4. * ui_scale)
            utils.draw_rounded_rect(shadow.org, shadow.size,
                                    button_corner_radius * ui_scale,
                                    color=RGBA(*color_shadow), sharpness=shadow_sharpness)

        utils.draw_rounded_rect(self.button.org, self.button.size,
                                button_corner_radius * ui_scale,
                                color=bg_color, sharpness=0.9)

        if self._outer_label:
            gl.glPushMatrix()
            glfont.push_state()
            gl.glTranslatef(self.field.org.x + button_text_padding * ui_scale,
                            self.field.org.y + button_text_padding * ui_scale,0)
            glfont.set_color_float(color_text_default)
            glfont.draw_limited_text(0,0,self._outer_label,self.field.size.x)
            glfont.pop_state()
            gl.glPopMatrix()

        gl.glPushMatrix()
        glfont.push_state()
        gl.glTranslatef(self.button.org.x + button_text_padding * ui_scale,
                        self.button.org.y + button_text_padding * ui_scale,0)
        glfont.set_color_float(self.text_color[:])
        glfont.draw_limited_text(x_spacer*ui_scale,0,self._label,self.button.size.x - 2. *  button_text_padding * ui_scale)
        glfont.pop_state()
        gl.glPopMatrix()

    cpdef handle_input(self,Input new_input,bint visible,bint parent_read_only = False):
        if not (self._read_only or parent_read_only):
            global should_redraw

            for b in new_input.buttons:
                if  visible and self.button.mouse_over(new_input.m):
                    if b[1] == 1:
                        self.selected = True
                        should_redraw = True
                if self.selected and b[1] == 0:
                    self.selected = False
                    should_redraw = True
                    self.function()

    @property
    def outer_label(self):
        return self._outer_label

    @outer_label.setter
    def outer_label(self,basestring val):
        if self._outer_label != val:
            self._outer_label = val
            global should_redraw
            should_redraw = True


cdef class Separator(UI_element):
    cdef float separator_height

    def __cinit__(self):
        self.outline = FitBox(Vec2(0,0),Vec2(0,0))

    def __init__(self):
        pass

    cpdef draw(self, FitBox parent,bint nested=True, bint parent_read_only = False):
        self.outline.compute(parent)

        line(Vec2(self.outline.org.x + outline_padding,
                  self.outline.org.y + outline_padding),
             Vec2(self.outline.org.x - outline_padding + self.outline.size.x,
                  self.outline.org.y + outline_padding),
             RGBA(*menu_line))

        self.outline.design_size.y = 1.5 * ui_scale * outline_padding * 2
        self.outline.compute(parent)

    cpdef precompute(self, FitBox parent):
        self.outline.compute(parent)
        self.outline.design_size.y = 1.5 * ui_scale * outline_padding * 2
        self.outline.compute(parent)

cdef class Info_Text(UI_element):
    cdef basestring _text
    cdef int max_height
    cdef FitBox text_area
    cdef float _text_size
    cdef RGBA _text_color
    cdef bint _centered

    def __cinit__(self, basestring text):
        self._text = text
        self.max_height = 200
        self.outline = FitBox(Vec2(0,0),Vec2(0,0))
        self.text_area = FitBox(Vec2(outline_padding,outline_padding),Vec2(-outline_padding,-outline_padding))
        self._text_size = size_text_info
        self._text_color = RGBA(*color_text_info)
        self._centered = False

    def __init__(self, basestring text):
        pass

    property text:
        def __get__(self):
            return self._text

        def __set__(self,basestring new_text):
            global should_redraw
            if self._text != new_text:
                should_redraw = True
                self._text = new_text

    property text_size:
        def __get__(self):
            return self._text_size

        def __set__(self, float new_text_size):
            global should_redraw
            if self._text_size != new_text_size:
                should_redraw = True
                self._text_size = new_text_size

    property text_color:
        def __get__(self):
            return self._text_color

        def __set__(self, RGBA new_text_color):
            global should_redraw
            if self._text_color != new_text_color:
                should_redraw = True
                self._text_color = new_text_color

    property centered:
        #TODO: Replace this property with full support for setting horizontal and vertical alignment
        def __get__(self):
            return self._centered

        def __set__(self, bint new_value):
            global should_redraw
            if self._centered != new_value:
                should_redraw = True
                self._centered = new_value

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        #update appearance
        self.outline.compute(parent)
        self.text_area.compute(self.outline)
        glfont.push_state()
        glfont.set_color_float(self.text_color.as_tuple())
        glfont.set_size(self.text_size*ui_scale)
        if self.centered:
            glfont.set_align(fs.FONS_ALIGN_CENTER | fs.FONS_ALIGN_TOP)
            text_origin = (self.text_area.center[0], self.text_area.org.y)
        else:
            text_origin = (self.text_area.org.x, self.text_area.org.y)
        left_word, height = glfont.draw_breaking_text(text_origin[0], text_origin[1], self._text, self.text_area.size.x,self.max_height )
        glfont.pop_state()
        self.text_area.design_size.y  = (height - self.text_area.org.y) / ui_scale
        self.outline.design_size.y = self.text_area.design_size.y + outline_padding * 2
        self.text_area.compute(self.outline)
        self.outline.compute(parent)

    cpdef precompute(self,FitBox parent):
        self.outline.compute(parent)
        self.text_area.compute(self.outline)
        glfont.push_state()
        glfont.set_size(self.text_size*ui_scale)
        left_word, height = glfont.compute_breaking_text(self.text_area.org.x, self.text_area.org.y, self._text, self.text_area.size.x,self.max_height )
        glfont.pop_state()
        self.text_area.design_size.y  = (height - self.text_area.org.y) / ui_scale
        self.outline.design_size.y = self.text_area.design_size.y + outline_padding * 2
        self.text_area.compute(self.outline)
        self.outline.compute(parent)


cdef class Color_Legend(UI_element):
    cdef basestring _text
    cdef FitBox text_area
    cdef float _text_size
    cdef RGBA _text_color
    cdef RGBA _line_color
    cdef int max_height

    def __cinit__(self, object line_color, basestring text):
        self._text = text
        self.max_height = 200
        self.outline = FitBox(Vec2(0,0),Vec2(0,0))
        self.text_area = FitBox(
            Vec2(outline_padding * 6, outline_padding),
            Vec2(-outline_padding, -outline_padding)
        )
        self._text_size = size_text_info
        self._text_color = RGBA(*color_text_info)
        self._line_color = RGBA(*line_color)

    def __init__(self, object line_color, basestring text):
        pass

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        self.outline.compute(parent)

        line(Vec2(self.outline.org.x + outline_padding,
                  self.outline.org.y + outline_padding),
             Vec2(self.outline.org.x + ui_scale * outline_padding * 4,
                  self.outline.org.y + outline_padding),
             self._line_color)

        self.text_area.compute(self.outline)

        glfont.push_state()
        glfont.set_color_float(self._text_color.as_tuple())
        glfont.set_size(self._text_size * ui_scale)
        glfont.set_align(fs.FONS_ALIGN_MIDDLE | fs.FONS_ALIGN_LEFT)
        left_word, height = glfont.draw_breaking_text(
            self.text_area.org.x,
            self.text_area.org.y,
            self._text,
            self.text_area.size.x,
            self.max_height,
        )
        glfont.pop_state()
        self.text_area.design_size.y  = (height - self.text_area.org.y) / ui_scale
        self.outline.design_size.y = self.text_area.design_size.y + outline_padding
        self.text_area.compute(self.outline)
        self.outline.compute(parent)

    cpdef precompute(self, FitBox parent):
        self.outline.compute(parent)
        self.text_area.compute(self.outline)
        glfont.push_state()
        glfont.set_size(self._text_size * ui_scale)
        glfont.set_align(fs.FONS_ALIGN_MIDDLE | fs.FONS_ALIGN_LEFT)
        left_word, height = glfont.compute_breaking_text(
            self.text_area.org.x,
            self.text_area.org.y,
            self._text,
            self.text_area.size.x,
            self.max_height
        )
        glfont.pop_state()
        self.text_area.design_size.y  = (height - self.text_area.org.y) / ui_scale
        self.outline.design_size.y = self.text_area.design_size.y + outline_padding
        self.text_area.compute(self.outline)
        self.outline.compute(parent)

########## Thumb ##########


cdef class Thumb(UI_element):
    '''
    Not a classical UI element. Use button instead.
    Thumb is a circular type of switch/button
    It can also display status info via a overlay
    '''
    cdef public FitBox button
    cdef bint selected
    cdef int on_val,off_val
    cdef float offset_x, offset_y, offset_size, label_line_height
    cdef public basestring label_font
    cdef Synced_Value sync_val
    cdef RGBA _on_color, _off_color
    cdef basestring _status_text
    cdef object hotkey, label_getter

    def __cinit__(
        self,
        str attribute_name,
        object attribute_context = None,
        on_val=True,
        off_val=False,
        label=None,
        label_font='roboto',
        label_offset_x=0,
        label_offset_y=0,
        label_offset_size=0,
        label_line_height=1.0,
        setter=None,
        getter=None,
        hotkey = None,
        on_color=thumb_color_on,
        off_color=thumb_color_off,
        label_getter=None
    ):
        self.uid = id(self)
        self._label = label_getter() if label_getter is not None else (label or attribute_name[0])
        self.label_font = label_font
        self.label_getter = label_getter
        self.offset_x = label_offset_x
        self.offset_y = label_offset_y
        self.offset_size = label_offset_size
        self.label_line_height = label_line_height
        self.sync_val = Synced_Value(attribute_name,attribute_context,getter,setter)
        self.on_val = on_val
        self.off_val = off_val
        self.outline = FitBox(Vec2(0,0),Vec2(thumb_outline_size,thumb_outline_size))
        self.button = FitBox(Vec2(thumb_outline_pad,thumb_outline_pad),Vec2(-thumb_outline_pad,-thumb_outline_pad))
        self.selected = False
        self._on_color = RGBA(*on_color)
        self._off_color = RGBA(*off_color)
        self.hotkey = hotkey
        self._status_text = ''

    def __init__(self, *args, **kwargs):
        pass

    property status_text:
        def __get__(self):
            return self._status_text
        def __set__(self,basestring new_status_text):
            if self._status_text != new_status_text:
                global should_redraw
                should_redraw = True
                self._status_text = new_status_text

    property on_color:
        def __get__(self):
            return self._on_color
        def __set__(self, RGBA new_color):
            if self._on_color != new_color:
                global should_redraw
                should_redraw = True
                self._on_color = new_color

    property off_color:
        def __get__(self):
            return self._off_color
        def __set__(self, RGBA new_color):
            if self._off_color != new_color:
                global should_redraw
                should_redraw = True
                self._off_color = new_color

    cpdef sync(self):
        if self.label_getter is not None:
            # only redraws if label_getter returns a new value
            self.label = self.label_getter()
        self.sync_val.sync()

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        #update appearance
        self.outline.compute(parent)
        self.button.compute(self.outline)
        cdef int stroke_width = int(12 * ui_scale)
        cdef int stroke_width_half = int(stroke_width * 0.5)
        cdef int shadow_stroke_width = int(24 * ui_scale)
        cdef RGBA icon_color

        if self.sync_val.value == self.on_val:
            icon_color = self.on_color
        elif self.selected:
            icon_color = self.on_color
        else:
            icon_color = self.off_color

        #utils.draw_circle(self.button.center,radius=int(min(self.button.size)*.7)+stroke_width_half, stroke_width = shadow_stroke_width, color=RGBA(*thumb_color_shadow),sharpness=thumb_button_shadow_sharpness)
        utils.draw_circle(self.button.center,radius=int(min(self.button.size)*.77), stroke_width = stroke_width, color=RGBA(*thumb_color_shadow),sharpness=thumb_button_shadow_sharpness)
        utils.draw_circle(self.button.center,radius=int(min(self.button.size)*.77), stroke_width = stroke_width, color=icon_color,sharpness=thumb_button_sharpness)

        if self.selected:
            self.selected = False
            global should_redraw
            should_redraw = True

        glfont.push_state()
        glfont.set_font(self.label_font)
        glfont.set_align(fs.FONS_ALIGN_MIDDLE | fs.FONS_ALIGN_CENTER)
        glfont.set_size(max(1,int(min(self.button.size)+self.offset_size*ui_scale)-thumb_font_padding*ui_scale))
        glfont.set_color_float((0,0,0,0.5))
        glfont.set_blur(10.5)
        cdef int text_x = self.button.center[0]+int(self.offset_x*ui_scale)
        cdef int text_y = self.button.center[1]+int(self.offset_y*ui_scale)
        glfont.draw_multi_line_text(text_x, text_y, self._label, self.label_line_height)
        glfont.set_blur(0.5)
        glfont.set_color_float(icon_color[:])
        glfont.draw_multi_line_text(text_x, text_y, self._label, self.label_line_height)
        glfont.pop_state()


        # draw status text.
        glfont.push_state()
        glfont.set_font('roboto')
        glfont.set_align(fs.FONS_ALIGN_MIDDLE | fs.FONS_ALIGN_LEFT)
        glfont.set_size( max(1, int( (min(self.button.size) )-thumb_font_padding*ui_scale)/2.) )
        glfont.set_color_float((0,0,0,1))
        glfont.set_blur(10.5)
        glfont.draw_multi_line_text(self.button.center[0]+self.button.size.x/2.,self.button.center[1],self._status_text)
        glfont.set_color_float(icon_color[:])
        glfont.set_blur(.1)
        glfont.draw_multi_line_text(self.button.center[0]+self.button.size.x/2.,self.button.center[1],self._status_text)
        glfont.pop_state()


    cpdef handle_input(self,Input new_input,bint visible,bint parent_read_only = False):
        if not (self._read_only or parent_read_only):
            global should_redraw

            for b in new_input.buttons[:]:#we need to make a copy for remove to work as desired:
                if visible and self.button.mouse_over(new_input.m):
                    if b[1] in (1,2):
                        self.selected = True
                        should_redraw = True
                        new_input.buttons.remove(b)
                if self.selected and b[1] in (1,2) and (self.sync_val.value == self.on_val):
                    self.sync_val.value = self.off_val
                if self.selected and b[1] in (1,2) and (self.sync_val.value == self.off_val):
                    self.sync_val.value = self.on_val


            if self.hotkey is not None:
                for c in new_input.chars:
                    if c == self.hotkey:
                        self.selected = True
                        should_redraw = True
                        if self.sync_val.value == self.on_val:
                            self.sync_val.value = self.off_val
                        elif self.sync_val.value == self.off_val:
                            self.sync_val.value = self.on_val
                        break

                for k in new_input.keys:
                    if k[2] in (1,2):  #keydown
                        if k[0] == self.hotkey:
                            self.selected = True
                            should_redraw = True
                            if self.sync_val.value == self.on_val:
                                self.sync_val.value = self.off_val
                            elif self.sync_val.value == self.off_val:
                                self.sync_val.value = self.on_val
                            break

cdef class Icon(Thumb):
    cdef float _indicator_start, _indicator_stop
    cdef basestring _tooltip
    cdef bint being_hovered

    def __cinit__(self, *args, **kwargs):
        self._indicator_start = 0.
        self._indicator_stop = 0.
        self._tooltip = ''
        self.being_hovered = False
        self.outline = FitBox(Vec2(0,0),Vec2(icon_outline_size, icon_outline_size))

    @property
    def tooltip(self):
        return self._tooltip

    @tooltip.setter
    def tooltip(self, val):
        if self._tooltip != val:
            global should_redraw_overlay
            should_redraw_overlay = True
            self._tooltip = val

    @property
    def indicator_start(self):
        return self._indicator_start

    @indicator_start.setter
    def indicator_start(self, val):
        assert isinstance(val, float), 'Indicator values are required to be floats'
        val %= 1.
        if self._indicator_start != val:
            global should_redraw_overlay
            should_redraw_overlay = True
        self._indicator_start = val

    @property
    def indicator_stop(self):
        return self._indicator_stop

    @indicator_stop.setter
    def indicator_stop(self, val):
        assert isinstance(val, float), 'Indicator values are required to be floats'
        val %= 1.
        if self._indicator_stop != val:
            global should_redraw_overlay
            should_redraw_overlay = True
        self._indicator_stop = val

    cpdef handle_input(self,Input new_input,bint visible,bint parent_read_only = False):
        unused = super(Icon, self).handle_input(new_input, visible, parent_read_only)
        global should_redraw_overlay
        cdef bint hovering = visible and self.button.mouse_over(new_input.m)

        if hovering != self.being_hovered:
            self.being_hovered = hovering
            should_redraw_overlay = True
        elif new_input.s.y > 0.:
            should_redraw_overlay = True
        return unused

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        #update appearance
        self.outline.compute(parent)
        self.button.compute(self.outline)
        cdef tuple icon_color, bg_color
        cdef float ref_size = min(self.button.size)

        icon_color = 0, 0, 0, 1
        if self.sync_val.value == self.on_val or self.selected:
            bg_alpha = 1.
        else:
            bg_alpha = .6

        if self.selected:
            self.selected = False
            global should_redraw
            should_redraw = True

        utils.draw_points([self.button.center], size=int(ref_size*.7), color=RGBA(1., 1., 1., .3), sharpness=0.7)
        utils.draw_points([self.button.center], size=int(ref_size*.7), color=RGBA(1., 1., 1., bg_alpha), sharpness=0.9)

        glfont.push_state()
        glfont.set_font(self.label_font)
        glfont.set_align(fs.FONS_ALIGN_MIDDLE | fs.FONS_ALIGN_CENTER)
        glfont.set_size(max(1,int(ref_size+self.offset_size*ui_scale)-icon_font_padding*ui_scale))
        glfont.set_color_float((*icon_color[:3], 0.3))
        glfont.set_blur(3)
        cdef int text_x = self.button.center[0]+int(self.offset_x*ui_scale)
        cdef int text_y = self.button.center[1]+int(self.offset_y*ui_scale)
        glfont.draw_multi_line_text(text_x, text_y, self._label, self.label_line_height)
        glfont.set_blur(0.5)
        glfont.set_color_float(icon_color)
        glfont.draw_multi_line_text(text_x, text_y, self._label, self.label_line_height)
        glfont.pop_state()

    cpdef draw_overlay(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        cdef basestring T = self.tooltip
        cdef float ref_size = min(self.button.size)
        if self.indicator_start != self.indicator_stop:
            utils.draw_progress(self.button.center, self.indicator_start,
                                self.indicator_stop, inner_radius=int(ref_size*.625),
                                outer_radius=int(ref_size*.9), color=RGBA(*icon_progress_color),
                                sharpness=0.9)

        if self._tooltip == '' or not self.being_hovered:
            return # only draw tooltip when set

        cdef float text_height = tooltip_text_size * ui_scale
        cdef float pad_x = 0.25*text_height
        cdef float pad_y = .5*pad_x
        cdef float tip_width = text_height + pad_x  # == 2* pad_y

        cdef float vert_loc = self.button.center[1]
        cdef float tip_loc_x = self.button.org.x + 10.*ui_scale
        cdef float text_loc_x = tip_loc_x - tip_width - pad_x
        glfont.push_state()
        glfont.set_font('opensans')
        glfont.set_align(fs.FONS_ALIGN_MIDDLE | fs.FONS_ALIGN_RIGHT)
        glfont.set_size(text_height)
        glfont.set_blur(.0)
        cdef float text_width = glfont.text_bounds(text_loc_x, vert_loc, T)

        utils.draw_tooltip((tip_loc_x, vert_loc), (text_width, text_height),
                           padding=(pad_x, pad_y), tooltip_color=RGBA(.8, .8, .8, .9),
                           sharpness=.9)

        # glfont.set_color_float((0., 0., 0., 0.3))
        # glfont.set_blur(3)
        # glfont.draw_text(text_loc_x, vert_loc, T)

        glfont.set_color_float((0., 0., 0., 8.))
        glfont.draw_text(text_loc_x, vert_loc, T)

        glfont.pop_state()


cdef class Hot_Key(UI_element):
    '''
    Just a hotkey. Not displayed.
    '''
    cdef int on_val,off_val
    cdef Synced_Value sync_val
    cdef public RGBA on_color,off_color
    cdef object hotkey

    def __cinit__(self,str attribute_name, object attribute_context = None, on_val=True, off_val=False, label=None, setter=None, getter=None, hotkey = None):
        self.uid = id(self)
        self._label = label or attribute_name
        self.sync_val = Synced_Value(attribute_name,attribute_context,getter,setter)
        self.on_val = on_val
        self.off_val = off_val
        self.hotkey = hotkey
        self.outline = FitBox(Vec2(0,0),Vec2(0,0)) # we dont use it but we need to have it.


    def __init__(self,str attribute_name, object attribute_context = None,label = None, on_val = True, off_val = False ,setter= None,getter= None, hotkey = None):
        pass

    cpdef sync(self):
        self.sync_val.sync()

    cpdef handle_input(self,Input new_input,bint visible,bint parent_read_only = False):
        global should_redraw
        if not (self._read_only or parent_read_only):
            if self.hotkey is not None:
                for c in new_input.chars:
                    if c == self.hotkey:
                        if self.sync_val.value == self.on_val:
                            self.sync_val.value = self.off_val
                            should_redraw = True
                        elif self.sync_val.value == self.off_val:
                            self.sync_val.value = self.on_val
                            should_redraw = True
                        break
                for k in new_input.keys:
                    if k[2]  in (1,2):  #keydown
                        if k[0] == self.hotkey:
                            if self.sync_val.value == self.on_val:
                                self.sync_val.value = self.off_val
                                should_redraw = True
                            elif self.sync_val.value == self.off_val:
                                self.sync_val.value = self.on_val
                                should_redraw = True
                            break



# cython: language_level=3

from libcpp.vector cimport vector

from pyglui.cygl.glew cimport *


cdef class RGBA:
    cdef public float r,g,b,a

cpdef init()

cpdef draw_bars(verts, float height, float thickness=*,RGBA color=*)
cpdef draw_x(verts, float width, float height, float thickness=*,RGBA color=*)
cpdef draw_polyline(verts,float thickness=*,RGBA color=*,line_type =*)
cpdef draw_polyline_norm(verts,float thickness=*,RGBA color=*,line_type =*)

cpdef draw_points(points,float size=*,RGBA color=*,float sharpness=*)
cpdef draw_progress(location, float start, float stop, float inner_radius=*,
                    float outer_radius=*, RGBA color=*, float sharpness=*)
cdef draw_tooltip(tip_location, text_size, padding=*,
                   RGBA tooltip_color=*, float sharpness=*)
cpdef draw_circle(center_position=*,float radius=*,float stroke_width=*,
                  RGBA color=*,float sharpness=*)
cpdef draw_points_norm(points,float size=*,RGBA color=*,float sharpness=*)

cpdef draw_rounded_rect(origin, size, float corner_radius, RGBA color=*, float sharpness=*)

ctypedef struct fbo_tex_id:
    GLuint fbo_id
    GLuint tex_id

cdef class Render_Target:
    cdef fbo_tex_id fbo_tex

cdef class Named_Texture:
    cdef GLuint texture_id
    cdef bint use_yuv_shader

cdef class Sphere:
    cdef GLuint vertex_buffer_id
    cdef GLuint index_buffer_id
    cdef int vertex_buffer_size
    cdef int index_buffer_size
    cdef int indices_amount

cpdef push_ortho(int w,int h)
cpdef pop_ortho()



cdef class Timeline(UI_element):
    cdef public RGBA color
    cdef public double point_size, xstart, xstop, ystart, ystop, ypad
    cdef object draw_data, draw_label
    cdef FitBox data_area

    def __cinit__(self, basestring label, object draw_data_callback,
                  object draw_label_callback=None, double content_height=30.,
                  xpad=Vec2(130, 30), ypad=5., *args, **kwargs):
        self.uid = id(self)
        self.label = label
        self.draw_data = draw_data_callback
        self.draw_label = draw_label_callback
        self.ypad = ypad
        self.outline = FitBox(Vec2(0, 0.), Vec2(0, content_height + 2*self.ypad))
        self.data_area = FitBox(Vec2(xpad[0], self.ypad), Vec2(-xpad[1], -self.ypad))

    def __init__(self, *args, **kwargs):
        pass

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        self.outline.compute(parent)
        self.data_area.compute(self.outline)

        cdef int width, height
        width = int(self.data_area.size.x)
        height = int(self.data_area.size.y)
        if width < 1 or height < 1:
            return

        cdef int vp[4]  # current view port data
        gl.glGetIntegerv(gl.GL_VIEWPORT, &vp[0])
        gl.glPushAttrib(gl.GL_VIEWPORT_BIT)
        cdef int org_y = vp[3] - int(self.data_area.org.y) - int(self.data_area.size.y)

        # setup gl
        gl.glViewport(int(self.data_area.org.x), org_y, width, height)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        gl.glOrtho(0, width, height, 0, -1, 1)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPushMatrix()
        gl.glLoadIdentity()

        self.draw_data(width, height, ui_scale)

        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPopMatrix()
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPopMatrix()
        gl.glPopAttrib()

        width = int(self.data_area.org.x - self.outline.org.x - x_spacer * ui_scale)
        gl.glPushAttrib(gl.GL_VIEWPORT_BIT)
        gl.glViewport(int(self.outline.org.x), org_y, width, height)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        gl.glOrtho(0, width, height, 0, -1, 1)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPushMatrix()
        gl.glLoadIdentity()

        if self.draw_label is None:
            self.draw_label_default(width, height, ui_scale)
        else:
            self.draw_label(width, height, ui_scale)

        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPopMatrix()
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPopMatrix()
        gl.glPopAttrib()

        cdef tuple seperator = ((self.outline.org.x,
                                 self.outline.org.y + self.outline.size.y),
                                (self.outline.org.x + self.outline.size.x,
                                 self.outline.org.y + self.outline.size.y))
        utils.draw_polyline(seperator, color=RGBA(*color_line_default),
                            line_type=gl.GL_LINES, thickness=ui_scale)

    cpdef refresh(self):
        global should_redraw
        should_redraw = True

    @property
    def content_height(self):
        '''gets scale-independent timeline height excluding vertical padding'''
        return self.outline.design_size.y - 2 * self.ypad

    @content_height.setter
    def content_height(self, val):
        '''sets scale-independent timeline height excluding vertical padding'''
        if val != self.content_height:
            self.outline.design_size.y = val + 2 * self.ypad
            self.refresh()

    cpdef draw_label_default(self, width, height, scale):
        glfont.push_state()
        glfont.set_font('opensans')
        glfont.set_size(timeline_label_size * scale)
        glfont.set_blur(.1)
        glfont.set_color_float((1., 1., 1., .8))
        glfont.set_align(fs.FONS_ALIGN_TOP | fs.FONS_ALIGN_RIGHT)
        glfont.draw_text(width, 0, self.label)



import platform

import pytest


def test_import_numpy():
    import numpy


@pytest.mark.skipif(
    platform.system() == "Windows", reason="cysignals is not required on Windows"
)
def test_import_cysignals():
    import cysignals


def test_import_glfw():
    import glfw


def test_import_pyglui():
    import pyglui

    assert pyglui


def test_import_cygl():
    import pyglui.cygl


def test_import_cygl_shader():
    from pyglui.cygl import shader


def test_import_cygl_utils():
    from pyglui.cygl import utils


def test_import_ui():
    import pyglui.ui

    assert pyglui.ui





cdef float sort_key(UI_element elm):
        return elm._order


cdef class Base_Menu(UI_element):
    """
    Base class that other menu inherit from. Dont use this.
    This contains all methods/attributed shared by all derived menus.
    """

    cdef public list elements
    cdef FitBox element_space
    cdef readonly int header_pos_id
    cdef Draggable menu_bar,minimize_corner, resize_corner
    cdef public RGBA color

    cpdef sync(self):
        if self.element_space.has_area():
            for e in self.elements:
                e.sync()

    def append(self, obj):
        if not issubclass(obj.__class__,UI_element):
            raise Exception("Can only append UI elements, not: '%s'"%obj )
        self.elements.append(obj)
        global should_redraw
        should_redraw = True

    def insert(self,idx, obj):
        if not issubclass(obj.__class__,UI_element):
            raise Exception("Can only append UI elements, not: '%s'"%obj )
        self.elements.insert(idx,obj)
        global should_redraw
        should_redraw = True

    def extend(self, objs):
        for obj in objs:
            if not issubclass(obj.__class__,UI_element):
                raise Exception("Can only append UI elements, not: '%s'"%obj )
        self.elements.extend(objs)
        global should_redraw
        should_redraw = True

    def remove(self, obj):
        del self.elements[self.elements.index(obj)]
        global should_redraw
        should_redraw = True

    def __len__ (self):
        return len(self.elements)

    def __getitem__ (self,x):
        return self.elements[x]

    def __setitem__ (self,x,obj):
        self.elements[x] = obj
        global should_redraw
        should_redraw = True

    def __delitem__ (self,x):
        del self.elements[x]
        global should_redraw
        should_redraw = True

    def __contains__ (self,obj):
        return obj in self.elements

    def __bool__(self):
        return True


    cdef get_submenu_config(self):
        '''
        Menus are sometimes emebedded in Other menues. We load their configurations recursively.
        '''
        cdef dict submenus = {}
        for e in self.elements:
            if isinstance(e,(Growing_Menu, Scrolling_Menu, Stretching_Menu, Container)):
                submenus[e.label] = submenus.get(e.label,[]) + [e.configuration] #we could have two submenues with same label so we use a list for each submenu label cotaining the conf dicts for each menu
        return submenus

    cdef set_submenu_config(self,dict submenus):
        '''
        Growing menus are sometimes emebedded in Other menues. We save their configurations recursively.
        '''
        if submenus:
            for e in self.elements:
                if isinstance(e,(Growing_Menu,Scrolling_Menu,Stretching_Menu,Container)):
                    e.configuration = submenus.get(e.label,[{}]).pop(0) #pop of the first menu conf dict in the list.


    cdef draw_menu(self,bint nested):
        #draw translucent background
        cdef Vec2 tripple_h_size = Vec2(menu_move_corner_width*ui_scale,menu_move_corner_height*ui_scale)
        cdef Vec2 tripple_v_size = Vec2(menu_move_corner_height*ui_scale,menu_move_corner_width*ui_scale)
        cdef Vec2 menu_offset = Vec2(menu_pad*ui_scale,menu_pad*ui_scale)
        if nested:
            pass
        else:
            gl.glColor4f(self.color.r,self.color.g,self.color.b,self.color.a)

            if self.header_pos_id == 2: #left
                self.outline.org.x += menu_sidebar_pad*ui_scale
                self.outline.size.x -= menu_sidebar_pad*ui_scale
                self.outline.sketch(self.color)
                self.outline.org.x -= menu_sidebar_pad*ui_scale
                self.outline.size.x += menu_sidebar_pad*ui_scale

            elif self.header_pos_id == 3: #right
                self.outline.size.x -= menu_sidebar_pad*ui_scale
                self.outline.sketch(self.color)
                self.outline.size.x += menu_sidebar_pad*ui_scale
            else:
                self.outline.sketch(self.color)

        if self.resize_corner is not None:
            self.resize_corner.outline.compute(self.outline)
            tripple_d(self.resize_corner.outline.org,self.resize_corner.outline.size)

        if self.menu_bar is not None:
            self.menu_bar.outline.compute(self.outline)
            if self.minimize_corner is not None:
                self.minimize_corner.outline.compute(self.menu_bar.outline)
            #self.minimize_corner.outline.sketch()
            if 2 == self.header_pos_id: #left
                if self.element_space.has_area():
                    tripple_v(self.menu_bar.outline.org+menu_offset,tripple_v_size)
                # else:
                #     triangle_right(self.menu_bar.outline.org+menu_offset,tripple_v_size)
            elif 3 == self.header_pos_id: #right
                if self.element_space.has_area():
                    tripple_v(self.menu_bar.outline.org+menu_offset,tripple_v_size)
                else:
                    triangle_left(self.menu_bar.outline.org+menu_offset,tripple_v_size)
            elif 0 == self.header_pos_id:  #top (botton not implemented)
                if nested:
                    menu_offset.x = 0
                if self.element_space.has_area():
                    tripple_h(self.menu_bar.outline.org+menu_offset,tripple_h_size)
                else:
                    triangle_h(self.menu_bar.outline.org+menu_offset,tripple_h_size,RGBA(*color_line_default))

                glfont.draw_text(self.menu_bar.outline.org.x+menu_offset.x+menu_topbar_text_x_org*ui_scale,
                                 self.outline.org.y+menu_offset.y,self._label)
                line(Vec2(self.menu_bar.outline.org.x+menu_offset.x,self.menu_bar.outline.org.y+self.menu_bar.outline.size.y),
                     Vec2(self.menu_bar.outline.org.x+self.menu_bar.outline.size.x-menu_offset.x,self.menu_bar.outline.org.y+self.menu_bar.outline.size.y),
                     RGBA(*menu_line))
            elif isinstance(self, Timeline_Menu):
                size = Vec2(2. * timelines_draggable_size * ui_scale, timelines_draggable_size * ui_scale)
                org = Vec2(*self.menu_bar.outline.center) - size / 2.
                border_color = RGBA(*color_line_default)

                gl.glColor4f(border_color.r, border_color.g, border_color.b, border_color.a)

                line_w = 2. * ui_scale
                gl.glLineWidth(line_w)
                gl.glBegin(gl.GL_LINES)
                gl.glVertex3f(org.x, org.y + size.y / 4, 0)
                gl.glVertex3f(org.x + size.x, org.y + size.y / 4, 0)
                gl.glVertex3f(org.x, org.y + size.y / 2, 0)
                gl.glVertex3f(org.x + size.x, org.y + size.y / 2, 0)
                gl.glEnd()

                line_w *= 2
                gl.glLineWidth(line_w)
                gl.glBegin(gl.GL_LINES)
                gl.glVertex3f(self.element_space.org.x, self.element_space.org.y - line_w / 2, 0)
                gl.glVertex3f(self.element_space.org.x + self.element_space.size.x,
                              self.element_space.org.y - line_w / 2, 0)
                gl.glEnd()


            elif 5 == self.header_pos_id:  #headline
                if not self.collapsed:
                    glfont.draw_text(self.menu_bar.outline.org.x+menu_offset.x,
                                     self.outline.org.y+menu_offset.y,self._label)
                    line(Vec2(self.menu_bar.outline.org.x+menu_offset.x,self.menu_bar.outline.org.y+self.menu_bar.outline.size.y),
                         Vec2(self.menu_bar.outline.org.x+self.menu_bar.outline.size.x-menu_offset.x,self.menu_bar.outline.org.y+self.menu_bar.outline.size.y),
                         RGBA(*menu_line))

            # (botton not implemented)
    def collect_in_window(self,FitBox window):
        global should_redraw
        if self.outline.design_org.x > 0:
            self.outline.design_org.x = min(self.outline.design_org.x, window.size.x/ui_scale-100)
            should_redraw = True
        if self.outline.design_org.y > 0:
            self.outline.design_org.y = min(self.outline.design_org.y, window.size.y/ui_scale-100)
            should_redraw = True
        if self.outline.design_org.x < 0:
            self.outline.design_org.x = max(self.outline.design_org.x, -1* window.size.x/ui_scale-100)
            should_redraw = True
        if self.outline.design_org.y < 0:
            self.outline.design_org.y = min(self.outline.design_org.y, -1* window.size.y/ui_scale-100)
            should_redraw = True


cdef class Movable_Menu(Base_Menu):
    '''
    Abstract class that implemented movable ui elements.
    '''


    property header_pos:
        def __get__(self):
            header_pos_list = ['top','bottom','left','right','hidden','headline']
            return header_pos_list[self.header_pos_id]

        def __set__(self, header_pos):
            #if the menu position is user changable we want the dragable to catch the input. This is good of user interaction
            cdef bint catch_input =  bool(self.outline.design_org.x or self.outline.design_org.y)

            if header_pos == 'top':
                self.element_space = FitBox(Vec2(menu_pad,menu_topbar_pad),Vec2(-menu_pad,-menu_pad- menu_bottom_pad))
                self.menu_bar = Draggable(Vec2(0,0),Vec2(0,menu_topbar_pad),
                                            self.outline.design_org,
                                            arrest_axis=0,zero_crossing = False,
                                            catch_input = catch_input )
                self.minimize_corner = Draggable(Vec2(0,0),Vec2(menu_topbar_pad,0),
                                            self.outline.design_org,
                                            arrest_axis=0,zero_crossing = False,
                                            click_cb=self.toggle_iconified,catch_input = catch_input )
                if self.outline.design_size.x:
                    self.resize_corner = Draggable(Vec2(-resize_corner_size,-resize_corner_size),Vec2(0,0),
                                                self.outline.design_size,
                                                arrest_axis=0,zero_crossing = False)
                else:
                    self.resize_corner = None

            elif header_pos == 'bottom':
                self.element_space = FitBox(Vec2(menu_pad,menu_bottom_pad + menu_pad),Vec2(-menu_pad,-menu_pad- menu_topbar_pad))
                self.menu_bar = Draggable(Vec2(0,-menu_pad),Vec2(0,0),
                                            self.outline.design_size,
                                            arrest_axis=0,zero_crossing = False,
                                            catch_input = catch_input  )
                self.minimize_corner = Draggable(Vec2(0,0),Vec2(menu_topbar_pad,0),
                                            self.outline.design_org,
                                            arrest_axis=0,zero_crossing = False,
                                            click_cb=self.toggle_iconified,catch_input = catch_input )
                self.resize_corner = None

            elif header_pos == 'right':
                self.element_space = FitBox(Vec2(menu_pad,0),Vec2(-menu_pad-menu_sidebar_pad,0))
                self.menu_bar = Draggable(Vec2(-menu_sidebar_pad,0),Vec2(0,0),
                                            self.outline.design_size,
                                            arrest_axis=0,zero_crossing = False,
                                            catch_input = catch_input  )
                self.minimize_corner = Draggable(Vec2(0,0),Vec2(menu_topbar_pad,0),
                                            self.outline.design_org,
                                            arrest_axis=0,zero_crossing = False,
                                            click_cb=self.toggle_iconified,catch_input = catch_input )
                self.resize_corner = None

            elif header_pos == 'left':
                self.element_space = FitBox(Vec2(menu_sidebar_pad+menu_pad,0),Vec2(-menu_pad,0))
                self.menu_bar = Draggable(Vec2(0,0),Vec2(menu_sidebar_pad,0),
                                            self.outline.design_org,
                                            arrest_axis=0,zero_crossing = False,
                                            catch_input = catch_input  )
                self.minimize_corner = Draggable(Vec2(0,0),Vec2(0,menu_topbar_pad),
                                                 self.outline.design_org, arrest_axis=0,
                                                 zero_crossing = False, catch_input = catch_input)
                # remove click_cb to prevent interaction conflict in new Pupil ui

                if self.outline.design_size.x > 0:
                    self.resize_corner = Draggable(Vec2(-resize_corner_size,-resize_corner_size),Vec2(0,0),
                                                    self.outline.design_size,
                                                    arrest_axis=2,zero_crossing = False)

            elif header_pos == 'hidden':
                self.element_space = FitBox(Vec2(0,0),Vec2(0,0))
                self.resize_corner = None
                self.menu_bar = None
            elif header_pos == 'headline':
                self.menu_bar = Draggable(Vec2(0,0),Vec2(0,menu_topbar_pad),
                                            self.outline.design_org,
                                            arrest_axis=1,zero_crossing = False,
                                            catch_input = catch_input  )
                self.element_space = FitBox(Vec2(0,menu_topbar_pad),Vec2(0,0))
                self.resize_corner = None

            else:
                raise Exception("Header Positon argument needs to be one of 'top,right,left,bottom','headline' was %s "%header_pos)

            self.header_pos_id = ['top','bottom','left','right','hidden','headline'].index(header_pos)




cdef class Stretching_Menu(Base_Menu):
    '''
    A simple menu
    Not movable and fixed in height.
    It will space its content evenenly in y.
    '''
    cdef public bint collapsed

    def __cinit__(self,label,pos=(0,0),size=(200,100)):
        self.uid = id(self)
        self._label = label
        self.outline = FitBox(position=Vec2(*pos),size=Vec2(*size),min_size=Vec2(0,0))
        self.element_space = FitBox(position=Vec2(menu_pad,menu_pad),size=Vec2(-menu_pad,-menu_pad))
        self.elements = []
        self.color = RGBA(*rect_color_default)
        self.collapsed = False

    def __init__(self,label,pos=(0,0),size=(200,100)):
        pass

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only=False):
        cdef float h = 0,y_spacing=0,org_y=0
        if not self.collapsed:
            self.elements.sort(key=sort_key)
            self.outline.compute(parent)
            self.element_space.compute(self.outline)

            for e in self.elements:
                e.precompute(self.element_space)
                h += e.height

            y_spacing  = (self.element_space.size.y-h)/(len(self.elements)+1)
            org_y = self.element_space.org.y
            #if elements are not visible, no need to draw them.
            if self.element_space.has_area():
                self.element_space.org.y+= y_spacing
                for e in self.elements:
                    e.draw(self.element_space,nested= False, parent_read_only = parent_read_only or self._read_only)
                    self.element_space.org.y+= e.height + y_spacing
                self.outline.org.y = org_y

    cpdef draw_overlay(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        if not self.collapsed and self.element_space.has_area():
            self.elements.sort(key=sort_key)
            for e in self.elements:
                (<UI_element>e).draw_overlay(self.element_space, nested=False,
                                             parent_read_only=parent_read_only or self._read_only)

    cpdef handle_input(self, Input new_input,bint visible,bint parent_read_only = False):
        #if elements are not visible, no need to interact with them.
        if self.element_space.has_area():
            for e in self.elements:
                e.handle_input(new_input, visible,self._read_only or parent_read_only)


    @property
    def configuration(self):
        cdef dict submenus = self.get_submenu_config()
        return {'pos':self.outline.design_org[:],'size':self.outline.design_size[:],'collapsed':self.collapsed,'submenus':submenus}

    @configuration.setter
    def configuration(self,new_conf):
        self.outline.design_org[:] = new_conf.get('pos',self.outline.design_org[:])
        self.outline.design_size[:] = new_conf.get('size',self.outline.design_size[:])
        self.collapsed = new_conf.get('collapsed',self.collapsed)
        self.set_submenu_config(new_conf.get('submenus',{}))

cdef class Horizontally_Stretching_Menu(Base_Menu):
    '''
    A simple menu
    Not movable and fixed in height.
    It will space its content evenenly in x.
    '''
    cdef public bint collapsed

    def __cinit__(self,label,pos=(0,0),size=(200,100)):
        self.uid = id(self)
        self._label = label
        self.outline = FitBox(position=Vec2(*pos),size=Vec2(*size),min_size=Vec2(0,0))
        self.element_space = FitBox(position=Vec2(menu_pad,menu_pad),size=Vec2(-menu_pad,-menu_pad))
        self.elements = []
        self.color = RGBA(*rect_color_default)
        self.collapsed = False

    def __init__(self,label,pos=(0,0),size=(200,100)):
        pass

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only=False):
        cdef float w=0, x_spacing=0, org_x=0
        if not self.collapsed:
            self.elements.sort(key=sort_key)
            self.outline.compute(parent)
            self.element_space.compute(self.outline)

            for e in self.elements:
                e.precompute(self.element_space)
                w += e.width

            x_spacing  = (self.element_space.size.x-w)/(len(self.elements)+1)
            org_x = self.element_space.org.x
            #if elements are not visible, no need to draw them.
            if self.element_space.has_area():
                self.element_space.org.x+= x_spacing
                for e in self.elements:
                    e.draw(self.element_space,nested= False, parent_read_only = parent_read_only or self._read_only)
                    self.element_space.org.x+= e.width + x_spacing
                self.outline.org.x = org_x

    cpdef draw_overlay(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        if not self.collapsed and self.element_space.has_area():
            self.elements.sort(key=sort_key)
            for e in self.elements:
                (<UI_element>e).draw_overlay(self.element_space, nested=False,
                                             parent_read_only=parent_read_only or self._read_only)

    cpdef handle_input(self, Input new_input,bint visible,bint parent_read_only = False):
        #if elements are not visible, no need to interact with them.
        if self.element_space.has_area():
            for e in self.elements:
                e.handle_input(new_input, visible,self._read_only or parent_read_only)


    @property
    def configuration(self):
        cdef dict submenus = self.get_submenu_config()
        return {'pos':self.outline.design_org[:],'size':self.outline.design_size[:],'collapsed':self.collapsed,'submenus':submenus}

    @configuration.setter
    def configuration(self,new_conf):
        self.outline.design_org[:] = new_conf.get('pos',self.outline.design_org[:])
        self.outline.design_size[:] = new_conf.get('size',self.outline.design_size[:])
        self.collapsed = new_conf.get('collapsed',self.collapsed)
        self.set_submenu_config(new_conf.get('submenus',{}))

cdef class Growing_Menu(Movable_Menu):
    '''
    Growing_Menu is a movable object on the canvas grows with its content

    size height will be ignored as it depend on the content.
    size positive -> size from self.org
    size 0 -> span into parent context and lock it like this. If you want it draggable use -.001 or .001
    size negative -> make the box to up to size pixels to the parent container.
    position negative -> align to the opposite side of context
    position 0  -> span into parent context and lock it like this. If you want it draggable use -.001 or .001


    '''
    cdef bint _collapsed


    def __cinit__(self,label,pos=(0,0),size=(0,0),header_pos = 'top'):
        self.uid = id(self)
        self._label = label
        #design height will be overwritten in draw.
        if header_pos in ('top','bottom'):
            min_size = Vec2(menu_topbar_min_width,menu_topbar_pad)
        elif header_pos in ('left','right'):
            min_size = Vec2(menu_sidebar_pad,menu_sidebar_min_height)
        else:
            min_size = Vec2(0,0)
        self.outline = FitBox(position=Vec2(*pos),size=Vec2(*size),min_size=min_size)

        self.elements = []
        self.header_pos = header_pos
        self.color = RGBA(*rect_color_default)

    def __init__(self,label, pos=(0,0),size=(0,0),header_pos = 'top'):
        pass

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only=False):
        self.outline.compute(parent)
        self.element_space.compute(self.outline)

        #here we compute the requred design height of this menu.
        self.outline.design_size.y  = self.height/ui_scale
        #now we correct for any changes induced by the content.
        self.outline.compute(parent)
        self.element_space.compute(self.outline)
        self.draw_menu(nested)

        cdef float org_y = self.element_space.org.y
        #if elements are not visible, no need to draw them.
        if self.element_space.has_area():
            self.elements.sort(key=sort_key)
            for e in self.elements:
                e.draw(self.element_space,parent_read_only = parent_read_only or self._read_only)
                self.element_space.org.y+= e.height
            self.outline.org.y = org_y

    cpdef draw_overlay(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        if self.element_space.has_area():
            self.elements.sort(key=sort_key)
            for e in self.elements:
                (<UI_element>e).draw_overlay(self.element_space, parent_read_only=parent_read_only or self._read_only)


    cpdef handle_input(self, Input new_input,bint visible,bint parent_read_only = False):
        if self.resize_corner is not None:
            self.resize_corner.handle_input(new_input,visible)
        if self.menu_bar is not None:
            self.menu_bar.handle_input(new_input,visible)
        if self.minimize_corner is not None:
            self.minimize_corner.handle_input(new_input,visible)


        #if elements are not visible, no need to interact with them.
        if self.element_space.has_area():
            for e in self.elements:
                e.handle_input(new_input, visible,self._read_only or parent_read_only)

    cpdef precompute(self,FitBox parent):
        #here we compute the requred design height of this menu.
        self.outline.design_size.y  = self.height/ui_scale
        self.outline.compute(parent)

    property height:
        def __get__(self):
            cdef float height = 0.0000001 #0 is magic and not meant when hight is computed to be 0.
            if self.header_pos_id == 5 and self.collapsed:
                pass
            else:
                #space from outline to element space at top
                height += self.element_space.design_org.y*ui_scale
            #space from element_space to outline at bottom
            height -= self.element_space.design_size.y*ui_scale #double neg
            if self._collapsed:
                #elemnt space is 0
                pass
            else:
                for e in self.elements:
                    e.precompute(self.element_space)
                height += sum([<float>e.height for e in self.elements])
            return height


    def toggle_iconified(self):
        self.collapsed = not self.collapsed

    property collapsed:
        def __get__(self):
            return self._collapsed
        def __set__(self,collapsed):
            if collapsed != self._collapsed:
                global should_redraw
                should_redraw = True
                self._collapsed = not self._collapsed


    @property
    def configuration(self):
        cdef dict submenus = self.get_submenu_config()
        return {'pos':self.outline.design_org[:],'size':[self.outline.design_size[0],0],'collapsed':self.collapsed,'submenus':submenus}

    @configuration.setter
    def configuration(self,new_conf):
        #load from configutation if avaible, else keep old setting
        self.outline.design_org[:] = new_conf.get('pos',self.outline.design_org[:])
        self.outline.design_size[:] = new_conf.get('size',self.outline.design_size[:])
        self.collapsed = new_conf.get('collapsed',self.collapsed)
        self.header_pos = self.header_pos #update layout
        self.set_submenu_config(new_conf.get('submenus',{}))




cdef class Scrolling_Menu(Movable_Menu):
    '''
    Scrolling_Menu is a movable object on the canvas that contains other elements
    and scrolls when they overflow the space.

    size positive -> size from self.org
    size 0 -> span into parent context and lock it like this. If you want it draggable use -.001 or .001
    size negative -> make the box to up to size pixels to the parent container.
    position negative -> align to the opposite side of context
    position 0  -> span into parent context and lock it like this. If you want it draggable use -.001 or .001

    '''
    cdef FitBox uncollapsed_outline
    cdef Draggable scrollbar
    cdef Vec2 scrollstate
    cdef float scroll_factor
    cdef bint _collapsed

    def __cinit__(self,label,pos=(100,100),size=(200,100),header_pos = 'top'):
        self.uid = id(self)
        self._label = label

        if header_pos in ('top', 'bottom', 'headline'):
            min_size = Vec2(menu_topbar_min_width,menu_topbar_pad)
        elif header_pos in ('right'):
            min_size = Vec2(menu_sidebar_pad,menu_sidebar_min_height)
        elif header_pos in ('left'):
            min_size = Vec2(menu_topbar_min_width,menu_sidebar_min_height)
        else:
            min_size = Vec2(0,0)
        self.outline = FitBox(position=Vec2(*pos),size=Vec2(*size),min_size=min_size)
        self.uncollapsed_outline = self.outline.copy()
        self._collapsed = False
        self.elements = []

        self.scrollstate = Vec2(0,0)
        self.scrollbar = Draggable(Vec2(0,0),Vec2(0,0),self.scrollstate,arrest_axis=1,zero_crossing=True)
        self.header_pos = header_pos
        self.color = RGBA(*rect_color_default)

    def __init__(self,label,pos=(100,100),size=(200,100),header_pos = 'top'):
        pass

    def __str__(self):
        return "Scrolling_Menu {}".format(self._label)

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only=False):
        self.outline.compute(parent)
        self.element_space.compute(self.outline)

        # collapse/inflate if needed
        if self.collapsed == self.element_space.has_area():
            if self.collapsed:
                self.uncollapsed_outline = self.outline.copy()
                self.outline.min_size = Vec2(0., 0.)
                self.outline.collapse()
            else:
                self.outline.min_size[:] = self.uncollapsed_outline.min_size[:]
                self.outline.inflate(self.uncollapsed_outline)
            self.outline.compute(parent)
            self.element_space.compute(self.outline)

        if self.header_pos_id == 2:
            nested = nested or self.collapsed
        self.draw_menu(nested)

        #if elements are not visible, no need to draw them.
        if self.element_space.has_area():
            self.elements.sort(key=sort_key)
            self.draw_scroll_window_elements(parent_read_only)

    cpdef draw_overlay(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        if self.element_space.has_area():
            self.elements.sort(key=sort_key)
            for e in self.elements:
                (<UI_element>e).draw_overlay(self.element_space, parent_read_only=parent_read_only or self._read_only)

    cdef draw_scroll_window_elements(self, bint parent_read_only):

        #compute scroll stack height.
        cdef float h = sum([e.height for e in self.elements])

        #If the scollbar is not active, make sure the content is not scrolled away:
        if not self.scrollbar.selected or 1:
            #self.scrollstate.y = clamp(self.scrollstate.y,min(-h,self.element_space.size.y-h),0)
            self.scrollstate.y = clamp(self.scrollstate.y*ui_scale,min(0,-h+self.element_space.size.y),0)/ui_scale

        self.push_scissor()
        self.scrollbar.outline.compute(self.element_space)
        self.element_space.org.y += self.scrollstate.y*ui_scale

        #now we compute h for real:
        cdef float e_h = 0
        h = 0
        #render elements
        for e in self.elements:
            e.draw(self.element_space,parent_read_only = parent_read_only or self._read_only)
            e_h = e.height
            h += e_h
            self.element_space.org.y+= e_h
        self.element_space.compute(self.outline)

        self.pop_scissor()

        cdef float scroll_factor = self.element_space.size.y/(h or 1)
        cdef float scroll_handle_offset = max(0,-self.scrollstate.y*ui_scale)*scroll_factor
        cdef float scroll_handle_size = self.element_space.size.y * scroll_factor

        #display that we have scrollable content
        if scroll_factor < 1:
            #self.element_space.size.x -= 10*ui_scale #shrink element space width to make room for scroll bar
            v_pad = 10*ui_scale
            start = self.element_space.org + self.element_space.size + Vec2(5*ui_scale,scroll_handle_offset-self.element_space.size.y+v_pad)
            end = self.element_space.org + self.element_space.size + Vec2(5*ui_scale,scroll_handle_offset+scroll_handle_size-self.element_space.size.y-v_pad)
            line(start,end,RGBA(*color_line_default))

    cdef push_scissor(self):
        # compute and set gl scissor
        gl.glPushAttrib(gl.GL_SCISSOR_BIT)
        gl.glEnable(gl.GL_SCISSOR_TEST)
        cdef int sb[4]
        global window_size
        gl.glGetIntegerv(gl.GL_SCISSOR_BOX,sb)
        sb[1] = window_size.y-sb[1]-sb[3] # y-flipped coord system
        #deal with nested scissors
        cdef float org_x = max(sb[0],self.element_space.org.x)
        cdef float size_x = min(sb[0]+sb[2],self.element_space.org.x+self.element_space.size.x)
        size_x = max(0,size_x-org_x)
        cdef float org_y = max(sb[1],self.element_space.org.y)
        cdef float size_y = min(sb[1]+sb[3],self.element_space.org.y+self.element_space.size.y)
        size_y = max(0,size_y-org_y)
        gl.glScissor(int(org_x),window_size.y-int(org_y)-int(size_y),int(size_x),int(size_y))

    cdef pop_scissor(self):
        #restore scissor state
        gl.glPopAttrib()


    cpdef handle_input(self, Input new_input,bint visible,bint parent_read_only = False):
        global should_redraw
        cdef bint mouse_over_menu = 0

        if self.resize_corner is not None:
            self.resize_corner.handle_input(new_input,visible)
        if self.minimize_corner is not None and not (2 == self.header_pos_id and not self.element_space.has_area()):
            self.minimize_corner.handle_input(new_input,visible)
        if self.menu_bar is not None:
            self.menu_bar.handle_input(new_input,visible)

        #if elements are not visible, no need to interact with them.
        if self.element_space.has_area():
            # let the elements know that the mouse should be ignored
            # if outside of the visible scroll section
            mouse_over_menu =  self.element_space.org.y <= new_input.m.y <= self.element_space.org.y+self.element_space.size.y
            mouse_over_menu = mouse_over_menu and visible
            for e in self.elements:
                e.handle_input(new_input, mouse_over_menu,self._read_only or parent_read_only)

            # handle scrollbar interaction after menu items
            # so grabbing a slider does not trigger scrolling
            #mouse:
            self.scrollbar.handle_input(new_input,visible)
            #scrollwheel:
            if new_input.s.y and visible and self.element_space.mouse_over(new_input.m):
                self.scrollstate.y += new_input.s.y * 3
                new_input.s.y = 0
                should_redraw = True


    def toggle_iconified(self):
        self.collapsed = not self.collapsed

    @property
    def collapsed(self):
        return self._collapsed

    @collapsed.setter
    def collapsed(self, should_collapse):
        if should_collapse != self.collapsed:
            global should_redraw
            should_redraw = True
            self._collapsed = should_collapse

    @property
    def configuration(self):
        cdef dict submenus = self.get_submenu_config()
        if self.collapsed:
            return {'label': self.label,
                    'pos':self.outline.design_org[:],
                    'size':self.outline.design_size[:],
                    'scrollstate':self.scrollstate[:],
                    'collapsed':True,
                    'min_size': self.outline.min_size[:],
                    'uncollapsed_min_size': self.uncollapsed_outline.min_size[:],
                    'uncollapsed_pos':self.uncollapsed_outline.design_org[:],
                    'uncollapsed_size':self.uncollapsed_outline.design_size[:],
                    'submenus':submenus}
        else:
            return {'pos':self.outline.design_org[:],'size':self.outline.design_size[:],
                    'collapsed':False,'submenus':submenus,'scrollstate':self.scrollstate[:],
                    'min_size': self.outline.min_size[:]}

    @configuration.setter
    def configuration(self, new_conf):
        self.outline.design_org[:] = new_conf.get('pos', self.outline.design_org[:])
        self.outline.design_size[:] = new_conf.get('size', self.outline.design_size[:])
        self.outline.min_size[:] = new_conf.get('min_size', None)

        if new_conf.get('collapsed', False):

            self.uncollapsed_outline.design_org[:] = new_conf.get('uncollapsed_pos',self.outline.design_org[:])
            self.uncollapsed_outline.design_size[:] = new_conf.get('uncollapsed_size',self.outline.design_size[:])
            self.uncollapsed_outline.min_size[:] = new_conf.get('uncollapsed_min_size',None)

        self.collapsed = new_conf.get('collapsed', False)
        self.header_pos = self.header_pos #update layout
        self.scrollstate[:] = new_conf.get('scrollstate',(0,0))
        self.set_submenu_config(new_conf.get('submenus',{}))


cdef class Container(Base_Menu):

    cdef public UI_element horizontal_constraint, vertical_constraint

    def __cinit__(self, pos=(0., 0.), size=(0., 0.), padding=(0., 0.)):
        self.outline = FitBox(Vec2(*pos), Vec2(*size))
        self.element_space = FitBox(Vec2(*padding), Vec2(0., 0.) - Vec2(*padding))
        self.elements = []
        self.horizontal_constraint = None
        self.vertical_constraint = None
        self.label = 'Container'

    def init(self, *args, **kwargs):
        pass

    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only=False):
        # compute constraints
        cdef FitBox copy = parent.computed_copy()
        self.outline.compute(copy)
        if self.horizontal_constraint is not None:
            self.horizontal_constraint.precompute(parent)
            if 0 <= copy.org.x < self.horizontal_constraint.outline.org.x < copy.org.x + copy.size.x:
                copy.size.x = self.horizontal_constraint.outline.org.x - copy.org.x
                if isinstance(self.horizontal_constraint, Base_Menu) and self.horizontal_constraint.header_pos_id == 2 and not self.horizontal_constraint.collapsed:  # left header
                    copy.size.x += menu_sidebar_pad * ui_scale
            elif 0 <= self.horizontal_constraint.outline.org.x + self.horizontal_constraint.outline.size.x <= copy.org.x:
                copy.size.x = self.horizontal_constraint.outline.org.x + self.horizontal_constraint.outline.size.x - copy.org.x
                copy.org.x = self.horizontal_constraint.outline.org.x + self.horizontal_constraint.outline.size.x

        self.outline.compute(copy)
        self.element_space.compute(self.outline)

        self.elements.sort(key=sort_key)
        if self.element_space.has_area():
            for e in self.elements:
                e.draw(self.element_space,nested= False, parent_read_only = parent_read_only or self._read_only)

    cpdef draw_overlay(self,FitBox parent,bint nested=True, bint parent_read_only = False):
        if self.element_space.has_area():
            self.elements.sort(key=sort_key)
            for e in self.elements:
                (<UI_element>e).draw_overlay(self.element_space, nested=False,
                                             parent_read_only=parent_read_only or self._read_only)

    cpdef handle_input(self, Input new_input,bint visible,bint parent_read_only = False):
        #if elements are not visible, no need to interact with them.
        if self.element_space.has_area():
            for e in self.elements:
                e.handle_input(new_input, visible,self._read_only or parent_read_only)

    @property
    def configuration(self):
        cdef dict submenus = self.get_submenu_config()
        return {'pos':self.outline.design_org[:],'size':self.outline.design_size[:],'submenus':submenus}

    @configuration.setter
    def configuration(self,new_conf):
        self.outline.design_org[:] = new_conf.get('pos',self.outline.design_org[:])
        self.outline.design_size[:] = new_conf.get('size',self.outline.design_size[:])
        self.set_submenu_config(new_conf.get('submenus',{}))


cdef class Timeline_Menu(Scrolling_Menu):
    cpdef draw(self,FitBox parent,bint nested=True, bint parent_read_only=False):
        self.outline.compute(parent)
        self.element_space.compute(self.outline)
        self.element_space.sketch(RGBA(0., 0., 0., 0.3))
        super(Timeline_Menu, self).draw(parent, nested, parent_read_only)

    @property
    def collapsed(self):
        return self.outline.design_org.y >= -menu_topbar_pad

    @collapsed.setter
    def collapsed(self,collapsed):
        if collapsed != self.collapsed:
            self.toggle_iconified()

    def toggle_iconified(self):
        global should_redraw
        should_redraw = True

        if self.collapsed:
            self.outline.inflate(self.uncollapsed_outline)
        else:
            self.uncollapsed_outline = self.outline.copy()
            self.outline.collapse()

    def append(self, obj):
        if len(self.elements) == 0:
            self.collapsed = False
        super(Timeline_Menu, self).append(obj)

    def insert(self, idx, obj):
        if len(self.elements) == 0:
            self.collapsed = False
        super(Timeline_Menu, self).insert(idx, obj)

    def extend(self, objs):
        if len(self.elements) == 0:
            self.collapsed = False
        super(Timeline_Menu, self).extend(objs)

    def remove(self, obj):
        super(Timeline_Menu, self).remove(obj)
        if len(self.elements) == 0:
            self.collapsed = True



import pytest


@pytest.fixture
def pyglui_ui_instance():
    import glfw

    from pyglui import cygl, ui

    glfw.ERROR_REPORTING = "raise"
    try:
        glfw_init_successful = glfw.init()
        assert glfw_init_successful, "Failed to initialize GLFW"
        glfw.window_hint(glfw.VISIBLE, glfw.FALSE)
        window = glfw.create_window(200, 200, "Test window", None, None)
    except glfw.GLFWError as err:
        # Skip if CI does not allow creating a window, e.g. on macOS:
        # > NSGL: Failed to find a suitable pixel format
        glfw.terminate()
        pytest.skip(str(err))

    glfw.make_context_current(window)  # required for GLEW init
    cygl.utils.init()
    global_ui_instance = ui.UI()
    yield global_ui_instance
    global_ui_instance.terminate()
    glfw.destroy_window(window)
    glfw.terminate()


@pytest.fixture
def attribute_context():
    return {"test": 5}


def test_Color_Legend(pyglui_ui_instance, attribute_context):
    import pyglui.ui

    black = (0.0, 0.0, 0.0, 1.0)
    pyglui.ui.Color_Legend(black, "test color")



import platform

from Cython.Build import cythonize
from setuptools import Extension, setup

if platform.system() == "Darwin":
    includes = ["/System/Library/Frameworks/OpenGL.framework/Versions/Current/Headers/"]
    f = "-framework"
    link_args = [f, "OpenGL"]
    libs = []
    compile_args = ["-D FONTSTASH_IMPLEMENTATION", "-D GLFONTSTASH_IMPLEMENTATION"]
elif platform.system() == "Windows":
    includes = []
    libs = ["OpenGL32"]
    link_args = []
    compile_args = [
        "/DFONTSTASH_IMPLEMENTATION",
        "/DGLFONTSTASH_IMPLEMENTATION",
    ]  # http://msdn.microsoft.com/de-de/library/hhzbb5c8.aspx
else:
    includes = [
        "/usr/include/GL",
    ]
    libs = ["GL"]
    link_args = []
    compile_args = ["-D FONTSTASH_IMPLEMENTATION", "-D GLFONTSTASH_IMPLEMENTATION"]

extensions = [
    Extension(
        name="fontstash",
        sources=["fontstash.pyx"],
        include_dirs=includes + ["fontstash/src"],
        libraries=libs,
        extra_link_args=link_args,
        extra_compile_args=compile_args,
    )
]

# this package will be compiled into a single.so file.
setup(
    name="pyfontstash",
    version="0.2",
    author="Moritz Kassner",
    license="MIT",
    description="OpenGL font rendering. This module can also be used as a submodule for other cython projects that want to use OpenGL.",
    ext_modules=cythonize(extensions),
)



import os
import pathlib
import platform
import sys
from stat import ST_MTIME

import numpy
import pkgconfig
from Cython.Build import cythonize
from setuptools import Extension, setup

includes = ["src/pyglui/cygl/", ".", numpy.get_include()]
glew_binaries = []
lib_dir = []
link_args = []
fontstash_compile_args = [
    "-D FONTSTASH_IMPLEMENTATION",
    "-D GLFONTSTASH_IMPLEMENTATION",
]


def get_cysignals_include():
    import cysignals

    return cysignals.__path__[0]


if platform.system() == "Darwin":
    # find glew irrespective of version
    pkgconfig_glew = pkgconfig.parse("glew")
    includes += pkgconfig_glew["include_dirs"]
    lib_dir += pkgconfig_glew["library_dirs"]
    libs = pkgconfig_glew["libraries"]
    extra_compile_args = ["-Wno-strict-aliasing", "-O2"]

    # OpenGL + cysignal headers
    includes += [
        "/System/Library/Frameworks/OpenGL.framework/Versions/Current/Headers/",
        "/Applications/Xcode.app/Contents/Developer/Platforms/MacOSX.platform/Developer/SDKs/MacOSX.sdk/System/Library/Frameworks/OpenGL.framework/Headers/",
        get_cysignals_include(),
    ]

    # find glew.h to generate glew.pxd later
    for include in pkgconfig_glew["include_dirs"]:
        for glew_header in pathlib.Path(include).rglob("glew.h"):
            # convert from Path to str
            glew_header = str(glew_header)
            break

elif platform.system() == "Linux":
    glew_header = "/usr/include/GL/glew.h"
    includes += ["/usr/include/GL", get_cysignals_include()]
    libs = ["GLEW", "GL"]  # GL needed for fonstash
    extra_compile_args = ["-Wno-strict-aliasing", "-O2"]
elif platform.system() == "Windows":
    glew_header = "src/pyglui/cygl/win_glew/gl/glew.h"
    includes += ["src/pyglui/cygl/win_glew"]
    libs = ["glew32", "OpenGL32"]
    lib_dir = ["src/pyglui/cygl/win_glew"]
    link_args = []
    gl_compile_args = []  # ['/DGL_GLEXT_PROTOTYPES']
    extra_compile_args = ["-O2"]
    fontstash_compile_args = [
        "/DFONTSTASH_IMPLEMENTATION",
        "/DGLFONTSTASH_IMPLEMENTATION",
    ]
else:
    raise Exception("Platform build not implemented.")

includes_fontstash = ["src/pyglui/pyfontstash/fontstash/src"]

extensions = [
    Extension(
        name="pyglui.ui",
        sources=["src/pyglui/ui.pyx"],
        include_dirs=includes + includes_fontstash,
        libraries=libs,
        library_dirs=lib_dir,
        extra_link_args=link_args,
        extra_compile_args=extra_compile_args,
        language="c++",
        define_macros=[("GL_SILENCE_DEPRECATION", "1")],
    ),
    Extension(
        name="pyglui.graph",
        sources=["src/pyglui/graph.pyx"],
        include_dirs=includes + includes_fontstash,
        libraries=libs,
        library_dirs=lib_dir,
        extra_link_args=link_args,
        extra_compile_args=extra_compile_args,
        language="c++",
        define_macros=[("GL_SILENCE_DEPRECATION", "1")],
    ),
    Extension(
        name="pyglui.cygl.utils",
        sources=["src/pyglui/cygl/utils.pyx"],
        include_dirs=includes,
        libraries=libs,
        library_dirs=lib_dir,
        extra_link_args=link_args,
        extra_compile_args=extra_compile_args,
        language="c++",
    ),
    Extension(
        name="pyglui.cygl.shader",
        sources=["src/pyglui/cygl/shader.pyx"],
        include_dirs=includes,
        libraries=libs,
        library_dirs=lib_dir,
        extra_link_args=link_args,
        extra_compile_args=extra_compile_args,
        language="c++",
    ),
    Extension(
        name="pyglui.pyfontstash.fontstash",
        sources=["src/pyglui/pyfontstash/fontstash.pyx"],
        include_dirs=includes + includes_fontstash,
        libraries=libs,
        library_dirs=lib_dir,
        extra_link_args=link_args,
        extra_compile_args=extra_compile_args + fontstash_compile_args,
        define_macros=[("GL_SILENCE_DEPRECATION", "1")],
    ),
]

should_cythonize = any(
    not pathlib.Path(sfile)
    .with_suffix(".cpp" if extension.language == "c++" else ".c")
    .exists()
    for extension in extensions
    for sfile in extension.sources
    if pathlib.Path(sfile).suffix == ".pyx"
)
if should_cythonize:
    # 1. generate additional cython files
    print(f"Generating glew.pxd based on {glew_header}")

    dir_glew_generator = pathlib.Path("scripts")
    dir_glew_destination = pathlib.Path("src", "pyglui", "cygl")

    print(f"Adding {dir_glew_generator} to sys.path")
    sys.path.append(str(dir_glew_generator))
    from glew_pxd import generate_pxd

    print(f"Removing {dir_glew_generator} from sys.path")
    sys.path.remove(str(dir_glew_generator))
    print(f"Writing glew.pxd to {dir_glew_destination}")
    generate_pxd(glew_header, dir_glew_destination)

    # 2. cythonize
    extensions = cythonize(extensions)
else:
    # Replace cython files with compiled c(++) sources
    # https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            sfile = pathlib.Path(sfile)
            path, ext = os.path.splitext(sfile)
            if sfile.suffix == ".pyx":
                ext = ".cpp" if extension.language == "c++" else ".c"
                sfile = sfile.with_suffix(ext)
            sources.append(str(sfile))
        extension.sources[:] = sources

setup(ext_modules=extensions)



import functools
import logging
import time

import glfw
import numpy as np
from OpenGL.GL import *

# create logger for the context of this function
logger = logging.getLogger(__name__)

import time

from pyglui import ui
from pyglui.cygl.shader import Shader
from pyglui.cygl.utils import RGBA, draw_points, init
from pyglui.pyfontstash import fontstash as fs

width, height = (1280, 720)


def basic_gl_setup():
    glEnable(GL_POINT_SPRITE)
    glEnable(GL_VERTEX_PROGRAM_POINT_SIZE)  # overwrite pointsize
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
    glEnable(GL_BLEND)
    glClearColor(0.8, 0.8, 0.8, 1.0)
    glEnable(GL_LINE_SMOOTH)
    # glEnable(GL_POINT_SMOOTH)
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST)
    glEnable(GL_POLYGON_SMOOTH)
    glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST)


def adjust_gl_view(w, h, window):
    """
    adjust view onto our scene.
    """
    print(w, h)
    glViewport(0, 0, int(w), int(h))
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glOrtho(0, w, h, 0, -1, 1)
    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()


_MARKER_CIRCLE_RGB_OUTER = (0.0, 0.0, 0.0)
_MARKER_CIRCLE_RGB_MIDDLE = (1.0, 1.0, 1.0)
_MARKER_CIRCLE_RGB_INNER = (0.0, 0.0, 0.0)
_MARKER_CIRCLE_RGB_FEEDBACK_INVALID = (0.8, 0.0, 0.0)
_MARKER_CIRCLE_RGB_FEEDBACK_VALID = (0.0, 0.8, 0.0)

_MARKER_CIRCLE_SIZE_OUTER = 60
_MARKER_CIRCLE_SIZE_MIDDLE = 38
_MARKER_CIRCLE_SIZE_INNER = 19
_MARKER_CIRCLE_SIZE_FEEDBACK = 3

_MARKER_CIRCLE_SHARPNESS_OUTER = 0.9
_MARKER_CIRCLE_SHARPNESS_MIDDLE = 0.8
_MARKER_CIRCLE_SHARPNESS_INNER = 0.55
_MARKER_CIRCLE_SHARPNESS_FEEDBACK = 0.5


@functools.lru_cache(4)  # 4 circles needed to draw calibration marker
def _circle_points_around_zero(radius: float, num_points: int) -> np.ndarray:
    t = np.linspace(0, 2 * np.pi, num_points, dtype=np.float64)
    t.shape = -1, 1
    points = np.hstack([np.cos(t), np.sin(t)])
    points *= radius
    return points


@functools.lru_cache(4)  # 4 circles needed to draw calibration marker
def _circle_points_offset(
    offset, radius: float, num_points: int, flat: bool = True
) -> np.ndarray:
    # NOTE: .copy() to avoid modifying the cached result
    points = _circle_points_around_zero(radius, num_points).copy()
    points[:, 0] += offset[0]
    points[:, 1] += offset[1]
    if flat:
        points.shape = -1
    return points


def _draw_circle_filled(screen_point, size: float, color: RGBA, num_points: int = 50):
    points = _circle_points_offset(
        screen_point, radius=size, num_points=num_points, flat=False
    )
    glColor4f(color.r, color.g, color.b, color.a)
    glEnableClientState(GL_VERTEX_ARRAY)
    glVertexPointer(2, GL_DOUBLE, 0, points)
    glDrawArrays(GL_POLYGON, 0, points.shape[0])


def _draw_circle_marker_polygon(position):

    r2 = 2
    screen_point = position

    # _draw_circle_filled(
    #     screen_point,
    #     size=_MARKER_CIRCLE_SIZE_OUTER * r2,
    #     color=RGBA(1.0, 0.0, 0.0, alpha),
    #     num_points=50,
    # )
    alpha = 1

    _draw_circle_filled(
        screen_point,
        size=_MARKER_CIRCLE_SIZE_OUTER * r2,
        color=RGBA(*_MARKER_CIRCLE_RGB_OUTER, alpha),
    )
    _draw_circle_filled(
        screen_point,
        size=_MARKER_CIRCLE_SIZE_MIDDLE * r2,
        color=RGBA(*_MARKER_CIRCLE_RGB_MIDDLE, alpha),
    )
    _draw_circle_filled(
        screen_point,
        size=_MARKER_CIRCLE_SIZE_INNER * r2,
        color=RGBA(*_MARKER_CIRCLE_RGB_INNER, alpha),
    )
    _draw_circle_filled(
        screen_point,
        size=_MARKER_CIRCLE_SIZE_FEEDBACK * r2,
        color=RGBA(*_MARKER_CIRCLE_RGB_FEEDBACK_VALID, alpha),
    )


def _draw_circle_marker_pointshader(position):

    r2 = 2 * 2
    screen_point = position
    alpha = 1

    draw_points(
        [screen_point],
        size=_MARKER_CIRCLE_SIZE_OUTER * r2,
        color=RGBA(*_MARKER_CIRCLE_RGB_OUTER, alpha),
        sharpness=_MARKER_CIRCLE_SHARPNESS_OUTER,
    )
    draw_points(
        [screen_point],
        size=_MARKER_CIRCLE_SIZE_MIDDLE * r2,
        color=RGBA(*_MARKER_CIRCLE_RGB_MIDDLE, alpha),
        sharpness=_MARKER_CIRCLE_SHARPNESS_MIDDLE,
    )
    draw_points(
        [screen_point],
        size=_MARKER_CIRCLE_SIZE_INNER * r2,
        color=RGBA(*_MARKER_CIRCLE_RGB_INNER, alpha),
        sharpness=_MARKER_CIRCLE_SHARPNESS_INNER,
    )
    draw_points(
        [screen_point],
        size=_MARKER_CIRCLE_SIZE_FEEDBACK * r2,
        color=RGBA(*_MARKER_CIRCLE_RGB_FEEDBACK_VALID, alpha),
        sharpness=_MARKER_CIRCLE_SHARPNESS_FEEDBACK,
    )


class Timer:
    def __init__(self):
        self._times = []

    def __enter__(self, *args, **kwargs):
        self.t0 = time.perf_counter()

    def __exit__(self, *args, **kwargs):
        self._times.append(time.perf_counter() - self.t0)
        del self.t0

    def result(self):
        import pandas as pd

        return pd.Series(self._times).describe()


def demo():
    global quit
    quit = False

    # Callback functions
    def on_resize(window, w, h):
        h = max(h, 1)
        w = max(w, 1)
        hdpi_factor = (
            glfw.get_framebuffer_size(window)[0] / glfw.get_window_size(window)[0]
        )
        w, h = w * hdpi_factor, h * hdpi_factor
        active_window = glfw.get_current_context()
        glfw.make_context_current(active_window)
        # norm_size = normalize((w,h),glfw.get_window_size(window))
        # fb_size = denormalize(norm_size,glfw.get_framebuffer_size(window))
        adjust_gl_view(w, h, window)
        glfw.make_context_current(active_window)

    def on_close(window):
        global quit
        quit = True
        logger.info("Process closing from window")

    # get glfw started
    glfw.init()

    window = glfw.create_window(width, height, "pyglui demo", None, None)
    if not window:
        exit()

    glfw.set_window_pos(window, 0, 0)
    # Register callbacks for the window
    glfw.set_window_size_callback(window, on_resize)
    glfw.set_window_close_callback(window, on_close)
    # test out new paste function

    glfw.make_context_current(window)
    init()
    basic_gl_setup()
    glfw.swap_interval(0)

    print(glGetString(GL_VERSION))

    on_resize(window, *glfw.get_window_size(window))

    t_pointshader = Timer()
    t_polygon = Timer()
    t_swap = Timer()
    t_poll = Timer()

    while not quit:

        with t_pointshader:
            _draw_circle_marker_pointshader((320, 320))
        with t_polygon:
            _draw_circle_marker_polygon((960, 320))

        with t_swap:
            glfw.swap_buffers(window)

        with t_poll:
            glfw.poll_events()
        # adjust_gl_view(1280,720,window)
        glClearColor(1.0, 1.0, 1.0, 1)
        glClear(GL_COLOR_BUFFER_BIT)

    print(f"{t_pointshader.result()=}")
    print(f"{t_polygon.result()=}")
    print(f"{t_swap.result()=}")
    print(f"{t_poll.result()=}")

    glfw.terminate()
    logger.debug("Process done")


if __name__ == "__main__":
    if 1:
        demo()
    else:
        import cProfile
        import os
        import subprocess

        cProfile.runctx("demo()", {}, locals(), "example.pstats")
        gprof2dot_loc = "gprof2dot.py"
        subprocess.call(
            "python "
            + gprof2dot_loc
            + " -f pstats example.pstats | dot -Tpng -o example_profile.png",
            shell=True,
        )
        print(
            "created cpu time graph for example. Please check out the png next to this."
        )



cimport pyglui.pyfontstash.cfontstash as fs

#expose some constansts when c imported
FONS_ALIGN_LEFT = fs.FONS_ALIGN_LEFT
FONS_ALIGN_CENTER = fs.FONS_ALIGN_CENTER
FONS_ALIGN_RIGHT = fs.FONS_ALIGN_RIGHT
FONS_ALIGN_TOP = fs.FONS_ALIGN_TOP
FONS_ALIGN_MIDDLE = fs.FONS_ALIGN_MIDDLE
FONS_ALIGN_BOTTOM = fs.FONS_ALIGN_BOTTOM
FONS_ALIGN_BASELINE = fs.FONS_ALIGN_BASELINE
FONS_ZERO_TOPLEFT = fs.FONS_ZERO_TOPLEFT
FONS_ZERO_BOTTOMLEFT = fs.FONS_ZERO_BOTTOMLEFT

cdef unicode _to_unicode(object s):
    if type(s) is unicode:
        return <unicode>s
    else:
        return (<bytes>s).decode('utf-8')

cdef bytes _to_utf8_bytes(object s):
    if type(s) is unicode:
        return (<unicode>s).encode('utf-8')
    else:
        return <bytes>s



cdef class Context:
    def __cinit__(self,atlas_size = (1024,1024),flags = fs.FONS_ZERO_TOPLEFT):
        self.ctx = fs.glfonsCreate(atlas_size[0],atlas_size[1],flags)

    def __init__(self,atlas_size = (1024,1024),flags = fs.FONS_ZERO_TOPLEFT):
        self.fonts = {}

    def __dealloc__(self):
        fs.glfonsDelete(self.ctx)

    def add_font(self, object name, object font_loc):
        cdef int font_id = fs.FONS_INVALID

        font_id = fs.fonsAddFont(self.ctx, _to_utf8_bytes(name), _to_utf8_bytes(font_loc))
        if font_id == fs.FONS_INVALID:
            raise Exception("Font could not be loaded from '%s'."%font_loc)
        else:
            self.fonts[name]=font_id

    property fonts:
        def __get__(self):
            return self.fonts

    def set_align(self,int align):
        '''
        bitwise or '|' the following:
        FONS_ALIGN_LEFT
        FONS_ALIGN_CENTER
        FONS_ALIGN_RIGHT
        FONS_ALIGN_TOP
        FONS_ALIGN_MIDDLE
        FONS_ALIGN_BOTTOM
        FONS_ALIGN_BASELINE
        '''
        fs.fonsSetAlign(self.ctx,align)

    def set_align_string(self,v_align='left',h_align='top'):
        v_align = {'left':FONS_ALIGN_LEFT,'center':FONS_ALIGN_CENTER,'right':FONS_ALIGN_RIGHT}[v_align]
        h_align = {'top':FONS_ALIGN_TOP,'middle':FONS_ALIGN_MIDDLE,'bottom':FONS_ALIGN_BOTTOM}[h_align]
        self.set_align(v_align | h_align)

    def set_blur(self,float blur):
        fs.fonsSetBlur(self.ctx,blur)

    def push_state(self):
        fs.fonsPushState(self.ctx)

    def pop_state(self):
        fs.fonsPopState(self.ctx)

    def clear_state(self):
        fs.fonsClearState(self.ctx)

    def set_size(self, float size):
        fs.fonsSetSize(self.ctx,size)

    def set_spacing(self, float spacing):
        fs.fonsSetSpacing(self.ctx,spacing)

    def set_font(self,object font_name):
        fs.fonsSetFont(self.ctx,self.fonts[font_name])

    def set_font_id(self,int font_id):
        fs.fonsSetFont(self.ctx,font_id)

    cpdef draw_text(self,float x,float y,object text):
        cdef float dx = fs.fonsDrawText(self.ctx,x,y,_to_utf8_bytes(text),NULL)
        return dx

    cpdef draw_limited_text(self, float x, float y, text, float width):
        '''
        draw text limited in width - it will cut off on the right hand side.
        '''
        cdef unicode utext = _to_unicode(text)
        if fs.fonsTextBounds(self.ctx, 0,0, _to_utf8_bytes(utext), NULL, NULL) <= width:
            #early exit it fits
            return self.draw_text(x,y,utext)

        if width <= 0:
            #early exit even the smallest char would not fit
            return self.draw_text(x,y,'')



        # start_text_clip = int(width/avg_char_width)
        cdef float avg_char_width = fs.fonsTextBounds(self.ctx, 0,0, 'o', NULL, NULL)
        cdef int max_idx = len(utext)
        cdef int idx = int(width/avg_char_width)
        cdef unicode clip = utext[:idx]
        cdef bint initial_guess_fits = fs.fonsTextBounds(self.ctx, 0,0, _to_utf8_bytes(clip), NULL, NULL) <= width


        if initial_guess_fits:
            #we add chars until it does not fit. then go back one
            while 0 <= idx <= max_idx:
                idx +=1
                clip = utext[:idx]
                if fs.fonsTextBounds(self.ctx, 0,0, _to_utf8_bytes(clip), NULL, NULL) > width:
                    idx -=1
                    break
        else:
            #we remove chars until it does fit.
            while 0 <= idx <= max_idx:
                idx -=1
                clip = utext[:idx]
                if fs.fonsTextBounds(self.ctx, 0,0, _to_utf8_bytes(clip), NULL, NULL) < width:
                    break


        utext = utext[:idx] #+ '..'
        return self.draw_text(x,y,utext)

    cpdef get_first_char_idx(self, object text, float width):
        '''
        get the clip index for a given width
        '''
        cdef unicode utext = _to_unicode(text)
        cdef int idx = len(utext)
        cdef unicode clip
        # reverse the text
        utext = utext[::-1]

        while idx:
            clip = utext[:idx]
            if fs.fonsTextBounds(self.ctx, 0,0, _to_utf8_bytes(clip), NULL, NULL) <= width:
                break
            idx -=1

        return len(utext)-idx

    cpdef draw_multi_line_text(self, float x, float y, object text, float line_height = 1):
        '''
        draw multiple lines of text delimited by "\n"
        '''
        cdef float asc = 0,des = 0,lineh = 0
        fs.fonsVertMetrics(self.ctx, &asc,&des,&lineh)
        line_height *= lineh
        lines = text.split('\n')
        for l in lines:
            fs.fonsDrawText(self.ctx,x,y,_to_utf8_bytes(l),NULL)
            y += line_height


    cpdef draw_breaking_text(self, float x, float y, object text, float width,float height,float line_height = 1):
        '''
        draw a string of text breaking at the bounds.
        '''

        cdef unicode utext = _to_unicode(text)

        # first we figure out the v space
        cdef float asc = 0,des = 0,lineh = 0
        fs.fonsVertMetrics(self.ctx, &asc,&des,&lineh)
        line_height *= lineh
        cdef float max_y = y + height - line_height

        # second we break the text into lines
        cdef unicode clip
        cdef list words = utext.split(' ')
        cdef int idx = 0, max_idx = len(words)

        # now we draw words
        while words:
            clip = ' '.join(words[:idx+1])
            if idx >= max_idx or fs.fonsTextBounds(self.ctx, 0,0, _to_utf8_bytes(clip), NULL, NULL) > width or u'\n' in clip[:-1]:
                clip = u' '.join(words[:idx])
                if clip and clip[-1] == u'\n':
                    clip = clip[:-1]
                fs.fonsDrawText(self.ctx,x,y,_to_utf8_bytes(clip),NULL)
                words = words[idx:]
                idx = 1 #always draw the first word.
                y += line_height
                if y > max_y:
                    break
            else:
                idx +=1

        return words,y

    cpdef compute_breaking_text(self, float x, float y, object text, float width,float height,float line_height = 1):
        '''
        draw a string of text breaking at the bounds.
        '''
        cdef unicode utext = _to_unicode(text)


        # first we figure out the v space
        cdef float asc = 0,des = 0,lineh = 0
        fs.fonsVertMetrics(self.ctx, &asc,&des,&lineh)
        line_height *= lineh
        cdef float max_y = y + height - line_height

        # second we break the text into lines
        cdef unicode clip
        cdef list words = utext.split(' ')
        cdef int idx = 1, max_idx = len(words)

        cdef bint first = True
        # now we draw words
        while words:
            clip = ' '.join(words[:idx])
            if idx > max_idx or fs.fonsTextBounds(self.ctx, 0,0, _to_utf8_bytes(clip), NULL, NULL) > width:
                if first == True:
                    idx = max(0,idx-1)
                    first = False
                else:
                    idx = max(0,idx-2)
                clip = u' '.join(words[:idx])
                words = words[idx:]
                y += line_height
                if y > max_y:
                    break
            idx +=1

        return words,y


    cpdef char_cumulative_width(self,float x, float y, object text):
        '''
        return a list with the cumulative width of each char in given text
        can be used as a map for positioning the caret
        or determining if a mouse position is close to a caret position
        '''
        cdef unicode utext = _to_unicode(text)
        # break the text string into chars
        cdef int total = 0
        cdef list running_sum = [0]
        for c in utext:
            total += self.text_bounds(x,y,c)
            running_sum.append(total)
        return running_sum


    def text_bounds(self,float x,float y, object text):
        '''
        get the width of a text
        '''
        cdef float width
        width = fs.fonsTextBounds(self.ctx,x,y,_to_utf8_bytes(text),NULL,NULL)
        return width

    #todo:
    #fonsLineBounds

    cpdef vertical_metrics(self):
        cdef float asc = 0,des = 0,lineh = 0
        fs.fonsVertMetrics(self.ctx, &asc,&des,&lineh)
        return asc,des,lineh


    cpdef set_color_float(self,tuple color):
        cdef unsigned int ir,ig,ib,ia,c
        ir = int(color[0]*255)
        ig = int(color[1]*255)
        ib = int(color[2]*255)
        ia = int(color[3]*255)
        c = fs.glfonsRGBA(ir,ig,ib,ia)
        fs.fonsSetColor(self.ctx,c)

    def draw_debug(self,float x,float y):
        fs.fonsDrawDebug(self.ctx,x,y)
